"""R tests adapter for Otter Assign"""

import nbformat
import re

from jinja2 import Template
from textwrap import indent
from typing import Optional, Union

from ..question_config import QuestionConfig
from ..tests_manager import AssignmentTestsManager
from ...test_files.abstract_test import TestCase, TestFile
from ...utils import get_source


# DON'T change this template or the regex that removes hidden tests will break in the R adapter!
OTTR_TEST_FILE_TEMPLATE = Template(
    """\
test = list(
  name = "{{ name }}",
  cases = list({% for tc in test_cases %}
    ottr::TestCase$new(
      hidden = {% if tc.hidden %}TRUE{% else %}FALSE{% endif %},
      name = {% if tc.name %}"{{ tc.name }}"{% else %}NA{% endif %},
      points = {% if tc.points is none %}NA{% else %}{{ tc.points }}{% endif %},{% if tc.success_message %}
      success_message = "{{ tc.success_message }}",{% endif %}{% if tc.failure_message %}
      failure_message = "{{ tc.failure_message }}",{% endif %}
      code = {
        {{ indent(tc.body, "        ").lstrip() }}{# lstrip so that the first line indent is correct #}
      }
    ){% if not loop.last %},{% endif %}{% endfor %}
  )
)"""
)
OTTR_TEST_FILE_TEMPLATE.globals["indent"] = indent


class RAssignmentTestsManager(AssignmentTestsManager):
    """
    A class for creating and managing test cases for an R assignment.
    """

    def read_test(self, cell: nbformat.NotebookNode, question: QuestionConfig):
        source = get_source(cell)

        if source[0].lstrip().startswith("#"):
            hidden = bool(re.search(r"\bhidden\b", source[0], flags=re.IGNORECASE))
        else:
            hidden = False

        test_start_line = 0 if hidden else -1

        config, maybe_test_start_line = self._parse_test_config(source)
        test_start_line = (
            maybe_test_start_line if maybe_test_start_line is not None else test_start_line
        )

        test_name = config.get("name", None)
        hidden = config.get("hidden", hidden)
        points = config.get("points", None)
        success_message = config.get("success_message", None)
        failure_message = config.get("failure_message", None)

        self._add_test_case(
            question,
            TestCase(
                test_name,
                "\n".join(source[test_start_line + 1 :]),
                hidden,
                points,
                success_message,
                failure_message,
            ),
        )

    @staticmethod
    def _resolve_test_file_points(
        total_points: Optional[Union[int, float, list[Optional[Union[int, float]]]]],
        test_cases: list[TestCase],
    ) -> list[TestCase]:
        return TestFile.resolve_test_file_points(total_points, test_cases)

    def _format_test(
        self,
        name: str,
        points: Optional[Union[int, float, list[Optional[Union[int, float]]]]],
        all_or_nothing: bool,
        test_cases: list[TestCase],
    ) -> str:
        template_data = {"name": name, "test_cases": test_cases}
        return OTTR_TEST_FILE_TEMPLATE.render(**template_data)



"""Exporters for Otter Export"""

from typing import Literal, Optional, Union

from .base_exporter import BaseExporter
from .via_html import PDFViaHTMLExporter
from .via_latex import PDFViaLatexExporter


ExporterType = Union[Literal["html"], Literal["latex"]]


def get_exporter(exporter_type: Optional[ExporterType] = None) -> type[BaseExporter]:
    """
    Returns the preferred exporter to use. Defaults to PDF via LaTeX exporter. Pass a string to
    ``exporter_type`` to override this behavior.

    Args:
        exporter_type (``"html" | "latex" | None``): a string identifying the type of exporter to
            use; can be one of ``["html", "latex"]``

    Returns:
        ``type[otter.export.exporters.base_exporter.BaseExporter]``: the exporter class
    """
    # throw an error if the nbconvert version is < 6
    import nbconvert

    if int(nbconvert.__version__.split(".")[0]) < 6:
        raise RuntimeError(
            f"Otter is only compatible with nbconvert>=6.0.0, found {nbconvert.__version__}"
        )

    if exporter_type is not None:
        exporter_type = exporter_type.lower()

        if exporter_type == "html":
            return PDFViaHTMLExporter

        elif exporter_type == "latex":
            return PDFViaLatexExporter

        else:
            raise ValueError(f"{exporter_type} is not a valid PDF exporter")

    return PDFViaLatexExporter



"""Autograding process internals for Otter-Grader"""

import json
import os
import pandas as pd
import zipfile

from contextlib import nullcontext
from glob import glob
from typing import Any, Union

from .autograder_config import AutograderConfig
from .runners import create_runner
from .utils import capture_run_output, OtterRuntimeError, print_output
from ... import logging
from ...utils import chdir, OTTER_CONFIG_FILENAME
from ...version import LOGO_WITH_VERSION


__all__ = ["AutograderConfig", "capture_run_output", "main"]


LOGGER = logging.get_logger(__name__)


def main(autograder_dir: str, *, otter_run: bool = False, **kwargs: Any):
    """
    Run the autograding process.

    Args:
        autograder_dir (``str``): the absolute path of the directory in which autograding is occurring
            (e.g. on Gradescope, this is ``/autograder``)
        otter_run (``bool``): whether this function is being invoked by Otter Run (i.e. without
            containerization)
        **kwargs: keyword arguments for updating autograder configurations; these values override
            anything present in ``otter_config.json``
    """
    import dill

    config_fp = os.path.join(autograder_dir, "source", OTTER_CONFIG_FILENAME)
    if os.path.isfile(config_fp):
        with open(config_fp, encoding="utf-8") as f:
            config = json.load(f)
    else:
        config = {}

    runner = create_runner(config, autograder_dir=autograder_dir, **kwargs)
    runner.ag_config.otter_run = otter_run

    ctx = nullcontext()
    if runner.ag_config.log_level is not None:
        ctx = logging.level_context(runner.ag_config.log_level)

    with ctx:
        LOGGER.debug(f"Detected containerized grading (T/F): {'F' if otter_run else 'T'}")
        LOGGER.debug(f"Config file path was resolved as: {config_fp}")
        LOGGER.debug(f"Autograder config was created: {runner.ag_config}")

        if runner.ag_config.logo:
            # U+8207 is an invisible non-whitespace character; this should prevent gradescope from
            # incorrectly left-stripping the whitespace at the beginning of the logo
            print_output(f"{chr(8207)}\n", LOGO_WITH_VERSION, "\n", sep="")

        abs_ag_path = os.path.abspath(runner.ag_config.autograder_dir)
        with chdir(abs_ag_path):
            try:
                if runner.ag_config.zips:
                    with chdir("./submission"):
                        zips = glob("*.zip")
                        if len(zips) > 1:
                            raise OtterRuntimeError(
                                "More than one zip file found in submission and 'zips' config is true"
                            )

                        with zipfile.ZipFile(zips[0]) as zf:
                            zf.extractall()

                runner.prepare_files()
                scores = runner.run()
                with open("results/results.pkl", "wb+") as f:
                    dill.dump(scores, f)

                output = scores.to_gradescope_dict(runner.ag_config)

            except OtterRuntimeError as e:
                output = {
                    "score": 0,
                    "stdout_visibility": "hidden",
                    "tests": [
                        {
                            "name": "Autograder Error",
                            "output": f"Otter encountered an error when grading this submission:\n\n{e}",
                        },
                    ],
                }
                raise e

            finally:
                if "output" in vars():
                    with open("./results/results.json", "w+") as f:
                        json.dump(output, f, indent=4)

        print_output("\n\n", end="")

        df = pd.DataFrame(output["tests"])

        if runner.ag_config.print_score and "score" in df.columns:
            total: Union[int, float]
            possible: Union[int, float]
            total, possible = df["score"].sum(), df["max_score"].sum()
            if "score" in output:
                total, possible = output["score"], runner.ag_config.points_possible or possible
            perc = total / possible * 100
            print_output(f"Total Score: {total:.3f} / {possible:.3f} ({perc:.3f}%)\n")

        if runner.ag_config.print_summary:
            pd.set_option("display.max_rows", None)  # print all rows of the dataframe
            if "output" in df.columns:
                df.drop(columns=["output"], inplace=True)
            if "visibility" in df.columns:
                df.drop(columns=["visibility"], inplace=True)
            if "status" in df.columns:
                df.drop(columns=["status"], inplace=True)

            print_output(df)



"""Solution removal for Otter Assign"""

import copy
import nbformat as nbf
import re

from .assignment import Assignment
from .r_adapter import solutions as r_solutions
from .utils import add_tag, has_tag, is_cell_type, remove_output, remove_tag
from ..utils import get_source


ANSWER_CELL_TAG = "otter_answer_cell"
BLOCK_PROMPT = "..."
OTTER_INCLUDE_TAG = "otter_include"
SOLUTION_CELL_TAG = "otter_assign_solution_cell"


def has_seed(cell: nbf.NotebookNode) -> bool:
    """
    Determine whether a cell contains a seed line (a line ending in ``# SEED``).

    Args:
        cell (``nbformat.NotebookNode``): the cell

    Returns:
        ``bool``: whether the cell has a seed line
    """
    if not is_cell_type(cell, "code"):
        return False
    source = get_source(cell)
    return bool(source) and any([l.strip().endswith("# SEED") for l in source])


def overwrite_seed_vars(nb: nbf.NotebookNode, seed_variable: str, seed: int) -> nbf.NotebookNode:
    """
    Overwrite any assignments of the variable named ``seed_variable`` with the value ``seed`` in a
    notebook.

    Args:
        nb (``nbformat.NotebookNode``): the notebook to edit
        seed_variable (``str``): the variable to look for
        seed (``int``): the value to set for ``seed_variable``

    Returns:
        ``nbformat.NotebookNode``: the notebook with the variable substitutions made
    """
    nb = copy.deepcopy(nb)
    for cell in nb["cells"]:
        source = get_source(cell)
        for i, line in enumerate(source):
            match = re.match(rf"(\s*){seed_variable}\s*(=|<-)\s*", line)
            if match:
                source[i] = match.group(1) + f"{seed_variable} {match.group(2)} {seed}"
        cell["source"] = "\n".join(source)
    return nb


solution_assignment_regex = re.compile(
    r"(\s*(?:[\w.]+(?=[^\w.])(?:\[['\"]?.*['\"]?\])*(?:,\s*)?)+\s*=).* ?# ?SOLUTION"
)


def solution_assignment_sub(match: re.Match[str]) -> str:
    """
    Substitutes the first matching group  with `` ...``
    """
    prefix = match.group(1)
    return prefix + " ..."


solution_line_regex = re.compile(r"(\s*).* ?# ?SOLUTION")


def solution_line_sub(match: re.Match[str]) -> str:
    """
    Substitutes the first matching group  with ``...``
    """
    prefix = match.group(1)
    return prefix + "..."


begin_solution_regex = re.compile(r"(\s*)# BEGIN SOLUTION( NO PROMPT)?")
skip_suffixes = ["# SOLUTION NO PROMPT", "# BEGIN PROMPT", "# END PROMPT", "# SEED"]

SUBSTITUTIONS = {
    "python": [
        (solution_assignment_regex, solution_assignment_sub),
        (solution_line_regex, solution_line_sub),
    ],
    "r": r_solutions.SUBSTITUTIONS,
}


def replace_solutions(lines: list[str], lang: str) -> list[str]:
    """
    Replace solutions in ``lines``.

    Args:
        lines (``list[str]``): solutions as a list of strings
        lang (``str``): the language of the code in ``lines``

    Returns:
        ``list[str]``: stripped version of lines without solutions
    """
    block_prompt: str
    if lang == "r":
        from .r_adapter.solutions import BLOCK_PROMPT as block_prompt
    else:
        block_prompt = globals()["BLOCK_PROMPT"]

    stripped = []
    solution = False
    for line in lines:

        # don't keep the line if it should be skipped
        if any(line.rstrip().endswith(s) for s in skip_suffixes):
            continue

        # don't keep the line if inside a solution block
        if solution and not line.rstrip().endswith("# END SOLUTION"):
            continue

        # process the end of a solution block
        if line.rstrip().endswith("# END SOLUTION"):
            assert solution, f"END SOLUTION without BEGIN SOLUTION in {lines}"
            solution = False
            continue

        # process the beginning of a solution block
        begin_solution = begin_solution_regex.match(line)
        if begin_solution:
            assert not solution, f"Nested BEGIN SOLUTION in {lines}"
            solution = True
            if not begin_solution.group(2):
                line = begin_solution.group(1) + block_prompt
            else:
                continue
        for exp, sub in SUBSTITUTIONS[lang]:
            m = exp.match(line)
            if m:
                line = sub(m)
                break

        stripped.append(line)

    assert not solution, f"BEGIN SOLUTION without END SOLUTION in {lines}"

    return stripped


def remove_ignored_lines(lines: list[str]) -> list[str]:
    """
    Remove ignored lines in ``lines``.

    Args:
        lines (``list[str]``): cell source as a list of strings

    Returns:
        ``list[str]``: stripped version of lines without ignored lines
    """
    ignore_suffix = "# IGNORE"
    stripped = []
    in_block = False
    for line in lines:

        # don't keep the line if it is ignored
        if line.rstrip().endswith(ignore_suffix):
            continue

        # don't keep the line if we're in an ignore block
        if in_block and not line.rstrip().endswith("# END IGNORE"):
            continue

        # process the end of an ignore block
        if line.rstrip().endswith("# END IGNORE"):
            assert in_block, f"END IGNORE without BEGIN IGNORE in {lines}"
            in_block = False
            continue

        # processing the beginning of an ignore block
        if re.match(r"\s*#\s*BEGIN\s*IGNORE\s*", line, flags=re.IGNORECASE):
            assert not in_block, f"Nested BEGIN IGNORE in {lines}"
            in_block = True
            continue

        stripped.append(line)

    assert not in_block, f"BEGIN IGNORE without END IGNORE in {lines}"

    return stripped


def strip_ignored_lines(nb: nbf.NotebookNode) -> nbf.NotebookNode:
    """
    Create a copy of a notebook with ignored lines stripped.

    Args:
        nb (``nbformat.NotebookNode``): the notebook to strip

    Returns:
        ``nbformat.NotebookNode``: a copy of ``nb`` with ignored line stripped
    """
    nb = copy.deepcopy(nb)
    for cell in nb["cells"]:
        cell["source"] = "\n".join(remove_ignored_lines(get_source(cell)))
    return nb


def strip_solutions_and_output(assignment: Assignment, nb: nbf.NotebookNode) -> nbf.NotebookNode:
    """
    Create a copy of a notebook with solutions and outputs stripped.

    Args:
        assignment (``otter.assign.assignment.Assignment``): the assignment config
        nb (``nbformat.NotebookNode``): the notebook to strip

    Returns:
        ``nbformat.NotebookNode``: a stripped copy of ``nb``
    """
    nb = copy.deepcopy(nb)

    del_md_solutions = []
    for i, cell in enumerate(nb["cells"]):
        if has_tag(cell, SOLUTION_CELL_TAG):
            if is_cell_type(cell, "code"):
                cell["source"] = "\n".join(replace_solutions(get_source(cell), assignment.lang))
            elif is_cell_type(cell, "markdown"):
                if has_tag(cell, OTTER_INCLUDE_TAG):
                    cell = remove_tag(cell, OTTER_INCLUDE_TAG)
                else:
                    del_md_solutions.append(i)
            nb["cells"][i] = add_tag(remove_tag(cell, SOLUTION_CELL_TAG), ANSWER_CELL_TAG)

    del_md_solutions.reverse()
    for i in del_md_solutions:
        del nb["cells"][i]

    # remove output from student version
    remove_output(nb)

    return nb



"""Utilities for Otter Run"""

from collections.abc import Iterable
from contextlib import contextmanager
from io import StringIO
from typing import Any, Optional


_output: Optional[StringIO] = None
"""a StringIO object to write output to instead of stdout when it is being captured"""


class OtterRuntimeError(RuntimeError):
    """
    A an error inheriting from ``RuntimeError`` for Otter to throw during a grading process.
    """


def write_blank_page_to_stare_at_before_you(path: str):
    """
    Write a PDF file with a single blank page at the specified path.

    Args:
        path (``str``): path at which to write the PDF
    """
    from pypdf import PdfWriter

    w = PdfWriter()
    w.add_blank_page(612, 792)
    w.write(path)


@contextmanager
def capture_run_output() -> Iterable[StringIO]:
    """
    A context manager for capturing anything that Otter Run would normally print to stdout. Yields
    an ``io.StringIO`` object that the output will be written to.

    Yields:
        ``io.StringIO``: where the output will be written to
    """
    global _output
    _output = StringIO()
    yield _output
    _output = None


def print_output(*args: Any, **kwargs: Any):
    """
    Print output for Otter Run. If output is being captured, this prints the output to the
    ``io.StringIO`` object it is being captured to, otherwise it prints to stdout. All arguments are
    passed to ``print``.
    """
    print(*args, **kwargs, file=_output)



"""Utilities for Otter Export exporters"""

import re


BEGIN_QUESTION_REGEX = r"<!--\s*BEGIN QUESTION\s*-->"
END_QUESTION_REGEX = r"<!--\s*END QUESTION\s*-->"
NEW_PAGE_REGEX = r"<!--\s*NEW PAGE\s*-->"
NEW_PAGE_MARKER = "#newpage"
NEW_PAGE_CELL_SOURCE = f"<!-- {NEW_PAGE_MARKER} -->"
NEW_PAGE_CLASS_NAME = "otter-page-break-after"


def has_begin(line: str) -> bool:
    """
    Returns whether a string contains a begin question comment

    A begin question comment is an HTML comment on a single line that denotes the beginning of an
    export block. The begin question comment looks like:

    .. code-block:: html

        <!-- BEGIN QUESTION -->

    Args:
        line (``str``): the line to search

    Returns:
        ``bool``: whether the line contains a substring matching the begin question regex
    """
    return bool(re.search(BEGIN_QUESTION_REGEX, line, flags=re.IGNORECASE))


def has_end(line: str) -> bool:
    """
    Returns whether a string contains an end question comment

    An end question comment is an HTML comment on a single line that denotes the end of an export
    block. The begin question comment looks like:

    .. code-block:: html

        <!-- END QUESTION -->

    Args:
        line (``str``): the line to search

    Returns:
        ``bool``: whether the line contains a substring matching the end question regex
    """
    return bool(re.search(END_QUESTION_REGEX, line, flags=re.IGNORECASE))


def sub_end_for_new_page(line: str) -> str:
    """
    Substitutes an end question comment for a newpage comment

    The end question HTML comment (cf. ``otter.export.filter.has_end``) is replaced with the following
    HTML comment to indicate a pagebreak in the LaTeX template.

    .. code-block:: html

        <!-- #newpage -->

    Args:
        line (``str``): the line to substitute in

    Returns:
        ``str``: the line with the end question match substituted for the newpage comment
    """
    return re.sub(END_QUESTION_REGEX, NEW_PAGE_CELL_SOURCE, line)



"""Runs Otter-Grader's autograding process"""

try:
    from otter.run.run_autograder import main as run_autograder
except ModuleNotFoundError as e:
    if "'otter'" in str(e):
        raise RuntimeError(
            "The 'otter' module could not be imported. This is usually caused by errors while building the "
            "grading image, so check the image build logs and include them when requesting "
            "support."
        )
    raise e


if __name__ == "__main__":
    run_autograder("{{ autograder_dir }}")



"""Functions for converting Assign-formatted R Markdown files to and from notebook objects"""

import jupytext
import nbformat as nbf
import os
import re

from copy import deepcopy

from ..solutions import ANSWER_CELL_TAG, SOLUTION_CELL_TAG
from ..utils import remove_tag
from ...utils import get_source, NBFORMAT_VERSION, NOTEBOOK_METADATA_KEY


HTML_COMMENT_START = "<!--"
HTML_COMMENT_END = "-->"
EXTRACT_COMMENT_REGEX = re.compile(rf"{HTML_COMMENT_START}\s*(#\s*[\w ]+)\s*{HTML_COMMENT_END}")
CONFIG_START_REGEX = re.compile(r"#\s+(ASSIGNMENT\s+CONFIG|(BEGIN|END)\s+\w+)", re.IGNORECASE)
YAML_COMMENT_CHAR = "#"


def read_as_notebook(rmd_path: str) -> nbf.NotebookNode:
    """
    Read an R Markdown file and convert it to a master notebook to be used with Otter Assign.

    Args:
        rmd_path (``pathlib.Path | str)``: the path to the master R Markdown file

    Returns:
        ``nbformat.NotebookNode``: the R Markdown file as a master notebook
    """
    with open(rmd_path) as f:
        lines = [l.strip("\n") for l in f.readlines()]

    new_lines = []
    in_comment = False
    in_solution_region, just_closed_solution_region = False, False
    has_prompt = False
    for i, l in enumerate(lines):
        # prevent excess whitespace in the student version of the notebook caused by the removal of
        # the lines containing the solution
        if just_closed_solution_region:
            just_closed_solution_region = False
            if l == "":
                continue

        if in_comment and l.strip() == HTML_COMMENT_END:
            new_lines.append("<!-- #endraw -->")
            in_comment = False

        elif l.startswith(HTML_COMMENT_START):
            if HTML_COMMENT_END in l:
                if CONFIG_START_REGEX.search(l):
                    if "begin" in l.lower() and "prompt" in l.lower():
                        has_prompt = True
                        if new_lines[len(new_lines) - 1].strip() == "":
                            new_lines.pop(len(new_lines) - 1)

                    if has_prompt:
                        if "begin" in l.lower() and "solution" in l.lower():
                            has_prompt = False
                            if new_lines[len(new_lines) - 1].strip() == "":
                                new_lines.pop(len(new_lines) - 1)

                        elif "end" in l.lower() and "prompt" not in l.lower():
                            has_prompt = False

                    new_lines.append("<!-- #raw -->")
                    new_lines.append(EXTRACT_COMMENT_REGEX.match(l).group(1))
                    new_lines.append("<!-- #endraw -->")

                else:
                    if l == """<!-- #region tags=["otter_assign_solution_cell"] -->""":
                        in_solution_region = True
                    elif in_solution_region and l == "<!-- #endregion -->":
                        in_solution_region, just_closed_solution_region = False, True

                    new_lines.append(l)

            elif l.strip() == HTML_COMMENT_START:
                if i + 1 < len(lines) and CONFIG_START_REGEX.match(lines[i + 1]):
                    new_lines.append("<!-- #raw -->")
                    in_comment = True

            else:
                new_lines.append(l)

        else:
            new_lines.append(l)

    if in_comment:
        raise ValueError("R Markdown file ends with an unclosed HTML comment")

    nb = jupytext.reads("\n".join(new_lines), "Rmd", as_version=NBFORMAT_VERSION)
    nb["metadata"]["kernelspec"] = {"language": "r"}

    return nb


def write_as_rmd(nb: nbf.NotebookNode, rmd_path: str, has_solutions: bool):
    """
    Write an autograder- or student-formatted R notebook as an R Markdown file.

    Args:
        nb (``nbformat.NotebookNode``): the notebook to write
        rmd_path (``str``): the path at which to write the R Markdown file
        has_solutions (``bool``): whether the provided notebook is an autograder notebook
    """
    if os.path.splitext(rmd_path)[1] != ".Rmd":
        raise ValueError("The provided path does not have the .Rmd extension")

    nb = deepcopy(nb)

    # prevent neighboring markdown cells from having two lines inserted between them in the student
    # notebook (resolves whitespace issues caused by the use of prompts for written questions)
    if not has_solutions:
        for i, cell in enumerate(nb["cells"]):
            # remove tags from all cells since they aren't needed for Rmd files
            if "tags" in cell["metadata"]:
                tags = cell["metadata"]["tags"]
                if SOLUTION_CELL_TAG in tags:
                    tags.remove(SOLUTION_CELL_TAG)
                if ANSWER_CELL_TAG in tags:
                    tags.remove(ANSWER_CELL_TAG)
                if not tags:
                    cell["metadata"].pop("tags")
            if (
                i < len(nb["cells"]) - 1
                and cell["cell_type"] == "markdown"
                and nb["cells"][i + 1]["cell_type"] == "markdown"
            ):
                cell["metadata"]["lines_to_next_cell"] = 0

    # add assignment name to Rmd metadata if necessary
    assignment_name = nb["metadata"].get(NOTEBOOK_METADATA_KEY, {}).get("assignment_name", None)
    if assignment_name:
        config_cell = nb["cells"][0]
        source = get_source(config_cell)
        source.insert(-1, f'assignment_name: "{assignment_name}"')
        config_cell["source"] = "\n".join(source)

    jupytext.write(nb, rmd_path)



"""
Plugin for using Google Sheets to override scores for test cases
"""

import json
import os
import pandas as pd
import tempfile

from ..abstract_plugin import AbstractOtterPlugin


class GoogleSheetsGradeOverride(AbstractOtterPlugin):
    """
    Otter plugin for overriding test case scores with values in a Google Sheet on Gradescope. Uses
    provided Google Service Account credentials to pull in the spreadsheet as a dataframe and edits
    test case scores by matching on the Gradescope assignment ID, student email, and test case name.

    Implements the ``during_generate``, ``before_grading``, and ``after_grading`` events. Make sure
    to list this plugin as ``otter.plugins.builtin.GoogleSheetsGradeOverride``, otherwise the
    ``during_generate`` event of this plugin will not work.

    The google sheet should have the following format:

    =============== ================== =========== ======== ==============
     Assignment ID   Email              Question    Points   PDF
    =============== ================== =========== ======== ==============
     123456          student@univ.edu   q1a - 1     1        false
    =============== ================== =========== ======== ==============

    ``Assignment ID`` should be the ID of the assignment on Gradescope, ``Email`` should be the email
    address corresponding to the student's Gradescope account, ``Question`` should be the name of
    the question, and ``Points`` should be the number of points that the student should be assigned.
    ``PDF`` should be ``false`` if the student's PDF should *not* be regenerated during this run of
    the autograder.
    """

    IMPORTABLE_NAME = "otter.plugins.builtin.GoogleSheetsGradeOverride"

    def _load_df(self):
        """
        Uses the Google Sheets API credentials stored in ``self.plugin_config`` to read in the
        sheet using ``pandas``.

        Returns:
            ``pandas.core.frame.DataFrame``: the sheet as a dataframe
        """
        import gspread

        try:
            oauth_json = self.plugin_config["service_account_credentials"]
            with tempfile.NamedTemporaryFile(mode="w+", suffix=".json") as ntf:
                json.dump(oauth_json, ntf)
                ntf.seek(0)

                gc = gspread.service_account(filename=ntf.name)

            sheet_url = self.plugin_config["sheet_url"]
            sheet = gc.open_by_url(sheet_url)
            worksheet = sheet.get_worksheet(0)
            data = worksheet.get_all_values()
            colnames = data.pop(0)

            self._df = pd.DataFrame(data, columns=colnames)
        except Exception as e:
            if self.plugin_config.get("catch_api_error", True):
                print(f"Error encountered while loading grade override sheet:\n{e}")
                self._df = pd.DataFrame(
                    columns=["Assignment ID", "Email", "Question", "Points", "PDF"]
                )
            else:
                raise e

    @property
    def df(self):
        """
        The grade override information dataframe
        """
        if not hasattr(self, "_df") or self._df is None:
            self._load_df()
        return self._df

    def after_grading(self, results):
        """
        Modifies the results of grading by pulling in the Google Sheet as a dataframe and updating
        the scores for the test cases found.

        Args:
            results (``otter.test_files.GradingResults``): the results of grading
        """
        if self.submission_metadata:
            df = self.df.copy()
            df = df[
                df["Email"].isin([user["email"] for user in self.submission_metadata["users"]])
                & (df["Assignment ID"] == str(self.submission_metadata["assignment"]["id"]))
            ]
            for _, row in df.iterrows():
                results.update_score(row["Question"], float(row["Points"]))

    def during_generate(self, otter_config, assignment):
        """
        Takes a path to Google Service Account credentials stored in this plugin's config as key
        ``credentials_json_path`` and extracts the data from that file into the plugin's config as key
        ``service_account_credentials``.

        Args:
            otter_config (``dict``): the parsed Otter configuration JSON file
            assignment (``otter.assign.assignment.Assignment``): the assignment configurations if
                Otter Assign is used
        """
        if assignment is not None:
            curr_dir = os.getcwd()
            os.chdir(assignment.master.parent)

        cfg_idx = [
            self.IMPORTABLE_NAME in c.keys() for c in otter_config["plugins"] if isinstance(c, dict)
        ].index(True)
        creds_path = otter_config["plugins"][cfg_idx][self.IMPORTABLE_NAME]["credentials_json_path"]
        with open(creds_path, encoding="utf-8") as f:
            creds = json.load(f)

        otter_config["plugins"][cfg_idx][self.IMPORTABLE_NAME][
            "service_account_credentials"
        ] = creds

        if assignment is not None:
            os.chdir(curr_dir)

    def before_grading(self, config):
        """
        Controls whether or not a PDF is generated by setting the ``token`` key of the grading config
        to ``None`` if any of the rows in the dataframe matching this (email, assignment ID) tuple
        indicate that a PDF should not be generated.

        Args:
            config (``dict``): the grading configurations
        """
        if self.submission_metadata:
            df = self.df.copy()
            df = df[
                df["Email"].isin([user["email"] for user in self.submission_metadata["users"]])
                & (df["Assignment ID"] == str(self.submission_metadata["assignment"]["id"]))
            ]

            for _, row in df.iterrows():
                generate_pdf = row["PDF"].strip().lower()
                if generate_pdf in ["n", "false", "f", "0"]:
                    config.token = None



"""
Builtin Otter plugins
"""

from .gmail_notifications import GmailNotifications
from .grade_override import GoogleSheetsGradeOverride
from .rate_limiting import RateLimiting


__all__ = ["GmailNotifications", "GoogleSheetsGradeOverride", "RateLimiting"]



"""Local grading of submissions in Docker containers for Otter-Grader"""

import os
import pathlib
import re

from glob import glob
from multiprocessing import Queue
from typing import Optional, Union

from .containers import launch_containers
from .utils import merge_scores_to_df, prune_images, SCORES_DICT_PERCENT_CORRECT_KEY
from .. import logging
from ..run import AutograderConfig
from ..utils import assert_path_exists


ALLOWED_EXTENSIONS = ["ipynb", "py", "Rmd", "R", "r", "zip"]
LOGGER = logging.get_logger(__name__)


def main(
    *,
    name: Optional[str] = None,
    paths: Optional[Union[list[str], tuple[str]]] = None,
    output_dir: str = "./",
    autograder: str = "./autograder.zip",
    containers: int = 4,
    ext: str = "ipynb",
    summaries: bool = False,
    no_kill: bool = False,
    image: str = "ubuntu:22.04",
    pdfs: bool = False,
    prune: bool = False,
    force: bool = False,
    timeout: Optional[int] = None,
    no_network: bool = False,
    debug: bool = False,
    result_queue: Optional["Queue[str]"] = None,
):
    """
    Run Otter Grade.

    Grades a directory of submissions in parallel Docker containers. Results are written as a CSV
    file called ``final_grades.csv`` in ``output_dir``. If ``pdfs`` is true, the PDFs generated
    inside the Docker containers are copied into a subdirectory of ``output_dir`` called
    ``submission_pdfs``.

    If ``prune`` is true, Otter's dangling grading images are pruned and the program exits.

    Args:
        name (``str``): an assignment name to use in the Docker image tag; must be specified unless
            ``prune`` is true
        paths (``list[str] | tuple[str]``): paths to submission files or directories of submissions
            for grading
        output_dir (``str``): path to directory where output should be written
        autograder (``str``): path to an Otter autograder configuration zip file
        containers (``int``): number of containers to run in parallel
        ext (``str``): the submission file extension (to be used in a glob pattern)
        no_kill (``bool``): whether to keep containers after grading is finished
        image (``str``): a Docker image to use as the base image for the grading image
        pdfs (``bool``): whether to copy notebook PDFs out of the containers
        prune (``bool``): whether to prune the grading images; if true, no grading is performed
        force (``bool``): whether to force-prune the images (do not ask for confirmation)
        timeout (``int | None``): an execution timeout in seconds for each container
        no_network (``bool``): whether to disable networking in the containers
        debug (``bool``): whether to run autograding in debug mode
        result_queue (``multiprocessing.Queue[str] | None``): the queue to store progress messages

    Returns:
        ``float | None``: the percentage scored by that submission if a single file was graded
            otherwise ``None``

    Raises:
        ``FileNotFoundError``: if a provided directory or file doesn't exist
        ``ValueError``: if an unsupported extension is passed to ``ext``
    """
    if prune:
        prune_images(force=force)
        return

    if name is None:
        raise ValueError("You must specify an assignment name")
    elif not re.match(r"^[\w\-.]+$", name):
        raise ValueError(
            "Assignment names may only contain letters, numbers, underscores, dashes, and periods"
        )

    if not isinstance(paths, tuple) and not isinstance(paths, list):
        raise TypeError("paths must be a tuple of valid paths")
    elif len(paths) == 0:
        raise ValueError("No paths specified")

    # check file paths
    assert_path_exists(
        [
            (output_dir, True),
            (autograder, False),
        ]
    )

    if ext not in ALLOWED_EXTENSIONS:
        raise ValueError(f"Invalid submission extension specified: {ext}")

    out = pathlib.Path(output_dir)
    try:
        if result_queue:
            logging.add_queue_handler(result_queue)

        LOGGER.info("Launching Docker containers")

        pattern = f"*.{ext}"
        submission_paths = []
        for path in paths:
            if os.path.isdir(path):
                submission_paths.extend(glob(os.path.join(path, pattern)))
            else:
                submission_paths.append(path)

        LOGGER.debug(f"Resolved submission paths: {submission_paths}")

        pdf_dir = out / "submission_pdfs" if pdfs else None

        scores = launch_containers(
            autograder,
            submission_paths,
            num_containers=containers,
            base_image=image,
            tag=name,
            no_kill=no_kill,
            pdf_dir=pdf_dir,
            timeout=timeout,
            network=not no_network,
            config=AutograderConfig(
                {
                    "zips": ext == "zip",
                    "pdf": pdfs,
                    "debug": debug,
                }
            ),
        )

        LOGGER.info("Combining grades and saving")
    finally:
        logging.remove_queue_handlers()

    # Merge scores to dataframe
    output_df = merge_scores_to_df(scores)

    # write to CSV file
    output_df.to_csv(out / "final_grades.csv", index=False)

    # write score summaries to files
    if summaries:
        grading_summary_path = out / "grading-summaries"
        if not grading_summary_path.exists():
            grading_summary_path.mkdir()

        for s in scores:
            nb_name, _ = os.path.splitext(s.file)
            with open(grading_summary_path / f"{nb_name}.txt", mode="w") as f:
                f.write(s.summary())

    # return percentage if a single file was graded
    if len(paths) == 1 and os.path.isfile(paths[0]):
        return output_df[SCORES_DICT_PERCENT_CORRECT_KEY][1]



"""Non-containerized single notebook grading for Otter-Grader"""

import dill
import json
import os
import pathlib
import shutil
import tempfile
import zipfile

from typing import Optional

from .run_autograder import AutograderConfig, capture_run_output, main as run_autograder_main
from ..test_files import GradingResults


__all__ = ["AutograderConfig", "capture_run_output", "main"]


def main(
    submission: str,
    *,
    autograder: str = "./autograder.zip",
    output_dir: str = "./",
    no_logo: bool = False,
    debug: bool = False,
    extra_submission_files: Optional[list[str]] = None,
) -> GradingResults:
    """
    Grades a single submission using the autograder configuration ``autograder`` without
    containerization.

    Creates a temporary directory in the user's system and replicates grading container structure.
    Calls the autograder and loads the pickled results object. **Note:** This does not run any setup
    or installation files, so the user's environment will need to have everything pre-installed.

    Args:
        submission (``str``): path to a submission to grade
        autograder (``str``): path to an Otter configuration zip file
        output_dir (``str | None``): directory at which to copy the results JSON file; if ``None``,
            the results JSON file is not copied
        no_logo (``bool``): whether to suppress the Otter logo from being printed to stdout
        debug (``bool``); whether to run in debug mode (without ignoring errors)
        extra_submission_files (``list[str] | None``): extra files to copy into the submission
            directory; this should really only be used internally by Otter, so use at your own risk

    Returns:
        ``otter.test_files.GradingResults``: the grading results object
    """
    dp = tempfile.mkdtemp()

    try:
        ag_dir = os.path.join(dp, "autograder")

        for subdir in ["source", "submission", "results"]:
            path = os.path.join(ag_dir, subdir)
            os.makedirs(path, exist_ok=True)

        with open(os.path.join(ag_dir, "submission_metadata.json"), "w+") as f:
            json.dump({}, f)

        ag_zip = zipfile.ZipFile(autograder)
        ag_zip.extractall(os.path.join(ag_dir, "source"))
        ag_zip.close()

        if os.path.splitext(submission)[1] == ".zip":
            subm_zip = zipfile.ZipFile(submission)
            subm_zip.extractall(os.path.join(ag_dir, "submission"))
            subm_zip.close()

        else:
            shutil.copy(submission, os.path.join(ag_dir, "submission"))

        for file in extra_submission_files or []:
            fp = pathlib.Path(os.path.join(ag_dir, file))
            # create any intermediate directories before copying the file
            fp.parents[0].mkdir(parents=True, exist_ok=True)
            shutil.copy(file, os.path.join(ag_dir, "submission", file))

        logo = not no_logo
        run_autograder_main(ag_dir, logo=logo, debug=debug, otter_run=True)

        results_path = os.path.join(ag_dir, "results", "results.json")
        if output_dir:
            shutil.copy(results_path, output_dir)

        results_pkl_path = os.path.join(ag_dir, "results", "results.pkl")
        with open(results_pkl_path, "rb") as f:
            results = dill.load(f)

    finally:
        shutil.rmtree(dp)

    return results



"""Question configurations for Otter Assign"""

import fica

from typing import Any, Optional, Union


class QuestionConfig(fica.Config):
    """
    Configurations for a question.
    """

    name: str = fica.Key(
        description="(required) the path to a requirements.txt file",
    )

    manual: bool = fica.Key(
        description="whether this is a manually-graded question",
        default=False,
    )

    points: Union[int, float, None] = fica.Key(
        description="how many points this question is worth; defaults to 1 internally",
        default=None,
    )

    check_cell: bool = fica.Key(
        description="whether to include a check cell after this question (for autograded "
        "questions only)",
        default=True,
    )

    export: bool = fica.Key(
        description="whether to force-include this question in the exported PDF",
        default=False,
    )

    all_or_nothing: bool = fica.Key(
        description="whether points should be assigned on an all-or-nothing basis",
        default=False,
    )

    def __init__(
        self,
        user_config: Optional[dict[str, Any]] = None,
        documentation_mode: bool = False,
        **kwargs,
    ):
        if user_config is None:
            user_config = {}
        if "name" not in user_config and not documentation_mode:
            raise ValueError(f"Question name not specified: {user_config}")

        super().__init__(user_config=user_config, documentation_mode=documentation_mode, **kwargs)



"""Otter Check command-line utility"""

import os

from typing import Optional

from .logs import EventType, LogEntry
from .notebook import OTTER_LOG_FILENAME
from .utils import list_test_files
from .. import logging
from ..execute import grade_notebook
from ..test_files import GradingResults


ALLOWED_EXTENSIONS = {".py", ".ipynb"}
LOGGER = logging.get_logger(__name__)


def _log_event(
    event_type: EventType,
    results: Optional[list[GradingResults]] = None,
    question: Optional[str] = None,
    success: bool = True,
    error: Optional[Exception] = None,
):
    """
    Logs an event

    Args:
        event_type (``otter.logs.EventType``): the type of event
        results (``list[otter.test_files.GradingResults]``): the
            results of any checks recorded by the entry
        question (``str | None``): the question name for this check
        success (``bool``): whether the operation was successful
        error (``Exception | None``): the exception thrown by the operation, if applicable
    """
    LOGGER.debug(f"Creating a LogEntry of type {event_type}")

    LogEntry(
        event_type, results=results, question=question, success=success, error=error
    ).flush_to_file(OTTER_LOG_FILENAME)

    LOGGER.debug(f"LogEntry created successfully")


def main(
    file: str,
    *,
    tests_path: str = "./tests",
    question: Optional[str] = None,
    seed: Optional[int] = None,
):
    """
    Runs Otter Check

    Args:
        file (``str``): path to the file to check
        tests_path (``str``): path to tests directory
        question (``str | None``): test name to run; ``None`` if all tests should be run
        seed (``int | None``): a seed to set before execution
    """
    try:
        if question:
            LOGGER.debug(f"Determining question '{question}' test path in directory '{tests_path}'")

            test_path = os.path.join(tests_path, question + ".py")
            if not os.path.isfile(test_path):
                raise FileNotFoundError(f"Test {question} does not exist")

            LOGGER.info(f"Found test file {test_path}")
            qs = [test_path]

        else:
            LOGGER.info(f"Searching for test files in tests directory")

            qs = list_test_files(tests_path)

            LOGGER.debug(f"Found test files: {', '.join(qs)}")

        LOGGER.debug(f"Checking for existence of submission file '{file}'")
        if not os.path.isfile(file):
            raise FileNotFoundError(f"{file} does not exist")

        ext = os.path.splitext(file)[1]
        LOGGER.debug(f"Found submission file extension: '{ext}'")

        if ext not in ALLOWED_EXTENSIONS:
            raise ValueError(
                f"Invalid extension for file '{ext}'; must be one of {ALLOWED_EXTENSIONS}"
            )

        script = ext == ".py"
        LOGGER.debug(f"Determined if submission is a Python script: {script}")

        LOGGER.debug(f"Seed value: {seed}")
        LOGGER.info("Grading submission")

        results = grade_notebook(
            file,
            tests_glob=qs,
            test_dir=tests_path,
            script=script,
            seed=seed,
        )

        percentage = results.total / results.possible
        LOGGER.debug(f"Determined score percentage: {percentage}")

        if percentage == 1:
            output = "All tests passed!"

        else:
            output = "\n".join(repr(test_file) for _, test_file in results.results.items())

        print(output)

    except Exception as e:
        _log_event(EventType.CHECK, success=False, error=e)
        raise e

    else:
        _log_event(EventType.CHECK, results=results)



"""Configurations for the autograder"""

import fica

from typing import Any, Literal, Optional, Union


class AutograderConfig(fica.Config):
    """
    Configurations for the autograder.
    """

    score_threshold: Optional[Union[int, float]] = fica.Key(
        description="a score threshold for pass-fail assignments",
        default=None,
    )

    points_possible: Optional[Union[int, float]] = fica.Key(
        description="a custom total score for the assignment; if unspecified the sum of question "
        "point values is used.",
        default=None,
    )

    show_stdout: bool = fica.Key(
        description="whether to display the autograding process stdout to students on Gradescope",
        default=False,
    )

    show_hidden: bool = fica.Key(
        description="whether to display the results of hidden tests to students on Gradescope",
        default=False,
    )

    show_all_public: bool = fica.Key(
        description="whether to display all test results if all tests are public tests",
        default=False,
    )

    seed: Optional[int] = fica.Key(
        description="a random seed for intercell seeding",
        default=None,
    )

    seed_variable: Optional[str] = fica.Key(
        description="a variable name to override with the seed",
        default=None,
    )

    grade_from_log: bool = fica.Key(
        description="whether to re-assemble the student's environment from the log rather than "
        "by re-executing their submission",
        default=False,
    )

    serialized_variables: Optional[dict[str, str]] = fica.Key(
        description="a mapping of variable names to type strings for validating a deserialized "
        "student environment",
        default=None,
    )

    pdf: bool = fica.Key(
        description="whether to generate a PDF of the notebook when not using Gradescope "
        "auto-upload",
        default=False,
    )

    token: Optional[str] = fica.Key(
        description="a Gradescope token for uploading a PDF of the notebook",
        default=None,
    )

    course_id: Optional[Union[int, str]] = fica.Key(
        description="a Gradescope course ID for uploading a PDF of the notebook",
        default=None,
    )

    assignment_id: Optional[Union[int, str]] = fica.Key(
        description="a Gradescope assignment ID for uploading a PDF of the notebook",
        default=None,
    )

    filtering: bool = fica.Key(
        description="whether the generated PDF should have cells filtered out",
        default=False,
    )

    pagebreaks: bool = fica.Key(
        description="whether the generated PDF should have pagebreaks between filtered sections",
        default=False,
    )

    debug: bool = fica.Key(
        description="whether to run the autograder in debug mode (without ignoring errors)",
        default=False,
    )

    autograder_dir: str = fica.Key(
        description="the directory in which autograding is taking place",
        default="/autograder",
    )

    lang: Union[Literal["python"], Literal["r"]] = fica.Key(
        description="the language of the assignment; one of {'python', 'r'}",
        default="python",
        validator=fica.validators.choice(["python", "r"]),
    )

    miniconda_path: str = fica.Key(
        description="the path to the mamba install directory",
        default="/root/miniforge3",
    )

    plugins: list[Union[str, dict[str, Any]]] = fica.Key(
        description="a list of plugin names and configuration details for grading",
        default=[],
    )

    logo: bool = fica.Key(
        description="whether to print the Otter logo to stdout",
        default=True,
    )

    print_summary: bool = fica.Key(
        description="whether to print the grading summary",
        default=True,
    )

    print_score: bool = fica.Key(
        description="whether to print out the submission score in the grading summary",
        default=True,
    )

    zips: bool = fica.Key(
        description="whether zip files are being graded",
        default=False,
    )

    log_level: Optional[int] = fica.Key(
        description="a log level for logging messages; any value suitable for "
        "``logging.Logger.setLevel``",
        default=None,
    )

    assignment_name: Optional[str] = fica.Key(
        description="a name for the assignment to ensure that students submit to the correct "
        "autograder",
        default=None,
    )

    warn_missing_pdf: bool = fica.Key(
        description="whether to add a 0-point public test to the Gradescope output to indicate "
        "to students whether a PDF was found/generated for this assignment",
        default=False,
    )

    force_public_test_summary: bool = fica.Key(
        description="whether to show a summary of public test case results when show_hidden is "
        "true",
        default=True,
    )

    submit_blank_pdf_on_export_failure: bool = fica.Key(
        description="whether to submit a blank PDF to the manual-grading Gradescope assignment "
        "if a PDF cannot be generated from the submission",
        default=False,
    )

    use_submission_pdf: bool = fica.Key(
        description="use the PDF in the submission zip file instead of exporting a new one; if "
        "no PDF is present, a new one is generated anyway; assumes there is only 1 PDF file "
        "in the submission",
        default=False,
    )

    pdf_via_html: bool = fica.Key(
        description="use the PDF via HTML exporter to export the submission PDF; ignored for Rmd "
        "submissions",
        default=False,
    )

    otter_run: bool = False
    """whether this autograder run is being run by Otter Run (i.e. without containerization)"""



"""Support for notebook metadata test files"""

import json

from .exception_test import ExceptionTestFile
from .ok_test import OKTestFile
from ..nbmeta_config import NBMetadataConfig
from ..utils import NOTEBOOK_METADATA_KEY


class _NotebookMetadataTestFileMixin:
    """
    A mixin for test file classes that read their test data from a notebook's metadata.
    """

    @classmethod
    def from_file(cls, path, test_name):
        """
        Parse a test from a Jupyter notebook's metadata and return an instance of the test file
        class.

        Args:
            path (``str``): the path to the notebook
            test_name (``str``): the name of the test to extract from the metadata

        Returns:
            ``_NotebookMetadataTestFileMixin``: the new test file object created from the given file
        """
        with open(path, encoding="utf-8") as f:
            nb = json.load(f)

        test_spec = nb["metadata"][NOTEBOOK_METADATA_KEY]["tests"]
        if test_name not in test_spec:
            raise ValueError(f"Test {test_name} not found")

        test_spec = test_spec[test_name]
        return cls.from_metadata(test_spec, path=path)

    @classmethod
    def from_nbmeta_config(cls, path: str, nbmeta_config: NBMetadataConfig, test_name: str):
        """
        Parse a test from an ``NBMetadataConfig`` and return an instance of the test file class.

        Args:
            path (``str``): the path to the notebook
            nbmeta_config (``NBMetadataConfig``): the config
            test_name (``str``): the name of the test to extract from the metadata

        Returns:
            ``_NotebookMetadataTestFileMixin``: the new test file object created from the given file
        """
        test_spec = nbmeta_config.tests
        if test_name not in test_spec:
            raise ValueError(f"Test {test_name} not found")

        test_spec = test_spec[test_name]
        return cls.from_metadata(test_spec, path=path)


class NotebookMetadataExceptionTestFile(ExceptionTestFile, _NotebookMetadataTestFileMixin):
    """
    A single notebook metadata test file for Otter.

    Tests are defined in the metadata of a jupyter notebook as a JSON object with the ``otter`` key.
    The tests themselves are assumed to be base-64-encoded compiled ``code`` objects from test files.

    .. code-block:: json

    {
        "metadata": {
            "otter": {
                "tests": {
                    "q1": ""
                }
            }
        }
    }
    """


class NotebookMetadataOKTestFile(OKTestFile, _NotebookMetadataTestFileMixin):
    """
    A single notebook metadata test file for Otter.

    Tests are defined in the metadata of a jupyter notebook as a JSON object with the ``otter`` key.
    The tests themselves are OK-formatted.

    .. code-block:: json

    {
        "metadata": {
            "otter": {
                "tests": {
                    "q1": {}
                }
            }
        }
    }
    """



"""Autograder configuration generator for Otter-Grader"""

import importlib.resources
import json
import os
import pathlib
import re
import yaml
import zipfile

from dataclasses import dataclass
from glob import glob
from jinja2 import Environment, PackageLoader
from typing import Any, Literal, Optional, TYPE_CHECKING, Union

from .token import APIClient
from .utils import merge_conda_environments, zip_folder
from ..plugins import PluginCollection
from ..run import AutograderConfig
from ..utils import dump_yaml, load_default_file, OTTER_CONFIG_FILENAME
from ..version import __version__


DEFAULT_PYTHON_VERSION = "3.12"
JINJA_ENV = Environment(loader=PackageLoader(__name__), keep_trailing_newline=True)
MINIFORGE_VERSION = "25.3.1-0"
OTTER_ENV_NAME = "otter-env"
OTTR_VERSION = "1.5.2"
TEMPLATE_DIR = importlib.resources.files(__name__) / "templates"


@dataclass
class CondaEnvironment:

    python_version: str

    is_r: bool

    requirements: list[str]

    overwrite_requirements: bool

    user_environment: Optional[dict[str, Any]]

    exclude_conda_defaults: bool

    def to_dict(self):
        environment: dict[str, Any] = {
            "name": OTTER_ENV_NAME,
            "channels": ["defaults", "conda-forge"],
            "dependencies": [
                f"python={self.python_version}",
                "pip",
                "jupyter_server",
            ],
        }

        if self.exclude_conda_defaults:
            environment["channels"].remove("defaults")

        if self.is_r:
            environment["channels"].append("r")
            environment["dependencies"].extend(
                [
                    "gcc_linux-64",
                    "gxx_linux-64",
                    "libgit2",
                    "libgomp",
                    "liblzma-devel",
                    "zlib-devel-amzn2-aarch64",
                    "zlib",
                    "r-base>=4.0.0",
                    "r-devtools",
                    "r-essentials",
                    "r-gert",
                    "r-rmarkdown",
                    "r-startup",
                    "r-stringi",
                    "r-testthat",
                    "r-usethis",
                    f"r-ottr=={OTTR_VERSION}",
                    "rpy2",
                ]
            )

        pip_deps = (
            self.requirements
            if self.overwrite_requirements
            else [
                f"otter-grader[grading,plugins]=={__version__}",
                *self.requirements,
            ]
        )

        environment["dependencies"].append({"pip": pip_deps})

        if self.user_environment:
            environment = merge_conda_environments(
                self.user_environment, environment, OTTER_ENV_NAME
            )

        return environment

    def to_str(self):
        return dump_yaml(self.to_dict(), indent=2)


COMMON_TEMPLATES = [
    "common/run_autograder",
    "common/run_otter.py",
]

LANGUAGE_BASED_CONFIGURATIONS = {
    "python": {
        "test_file_pattern": "*.py",
        "requirements_filename": "requirements.txt",
        "templates": [*COMMON_TEMPLATES, "python/setup.sh"],
    },
    "r": {
        "test_file_pattern": "*.[Rr]",
        "requirements_filename": "requirements.R",
        "templates": [*COMMON_TEMPLATES, "r/setup.sh"],
    },
}


def main(
    *,
    tests_dir: str = "./tests",
    output_path: str = "autograder.zip",
    config: Optional[str] = None,
    no_config: bool = False,
    lang: Optional[Union[Literal["python"], Literal["r"]]] = None,
    requirements: Optional[str] = None,
    no_requirements: bool = False,
    overwrite_requirements: bool = False,
    environment: Optional[str] = None,
    no_environment: bool = False,
    username: Optional[str] = None,
    password: Optional[str] = None,
    token: Optional[str] = None,
    files: Optional[list[str]] = None,
    assignment: Optional["Assignment"] = None,
    plugin_collection: Optional[PluginCollection] = None,
    python_version: Optional[str] = None,
    channel_priority_strict: bool = True,
    exclude_conda_defaults: bool = False,
):
    """
    Run Otter Generate.

    Args:
        tests_dir (``str``): path to directory of test files for this assignment
        output_path (``str``): the path at which to write the output zip file
        config (``str``): path to an Otter configuration JSON file
        no_config (``bool``): disables auto-inclusion of Otter config file at ./otter_config.json
        lang (``"python" | "r" | None``): the language of the assignment
        requirements (``str``): path to a Python or R requirements file for this assignment
        no_requirements (``bool``): disables auto-inclusion of requirements file at ./requirements.txt
        overwrite_requirements (``bool``): whether to overwrite the default requirements instead of
            adding to them
        environment (``str``): path to a conda environment file for this assignment
        no_environment (``bool``): whether ``./environment.yml`` should be automatically checked if
            ``environment`` is unspecified
        username (``str``): a username for Gradescope for generating a token
        password (``str``): a password for Gradescope for generating a token
        token (``str``): a token for Gradescope
        files (``list[str] | None``): list of file paths to add to the zip file
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
            if used with Otter Assign
        python_version (``str | None``): the version of Python to use (installed with conda)
        channel_priority_strict (``bool``): whether to set conda's channel_priority to strict in
            the ``setup.sh`` file
        exclude_conda_defaults (``bool``): whether to exclude conda's defaults channel in the
            generated ``environment.yml`` file

    Raises:
        ``FileNotFoundError``: if the specified Otter configuration JSON file could not be found
        ``ValueError``: if the configurations specify a Gradescope course ID or assignment ID but not
            both
    """
    if files is None:
        files = []

    # read in otter config
    if config is None and os.path.isfile(OTTER_CONFIG_FILENAME) and not no_config:
        config = OTTER_CONFIG_FILENAME

    if config is not None and not os.path.isfile(config):
        raise FileNotFoundError(f"Could not find otter configuration file {config}")

    if config:
        with open(config, encoding="utf-8") as f:
            otter_config = json.load(f)
    else:
        otter_config = {}

    # if an empty/null token is specified, delete it
    if "token" in otter_config and not otter_config["token"]:
        otter_config.pop("token")

    # ensure that a token is present if necessary
    if "token" not in otter_config and token is not None:
        otter_config["token"] = token

    elif (
        "token" not in otter_config
        and "course_id" in otter_config
        and "assignment_id" in otter_config
    ):
        client = APIClient()
        if username is not None and password is not None:
            client.log_in(username, password)
            token = client.token
        else:
            token = client.get_token()
        otter_config["token"] = token

    elif ("course_id" in otter_config) ^ ("assignment_id" in otter_config):
        raise ValueError(f"Otter config contains 'course_id' or 'assignment_id' but not both")

    ag_config = AutograderConfig(otter_config)

    # update language
    if lang is not None:
        ag_config.lang = lang
        otter_config["lang"] = lang

    if ag_config.lang not in LANGUAGE_BASED_CONFIGURATIONS.keys():
        raise ValueError(f"Invalid language specified: {ag_config.lang}")

    lang_config = LANGUAGE_BASED_CONFIGURATIONS[ag_config.lang]

    templates = {}
    for template in lang_config["templates"]:
        fn = os.path.basename(template)
        templates[fn] = JINJA_ENV.get_template(template)

    if python_version is not None:
        match = re.match(r"(\d+)\.(\d+)(\.\d+)?", python_version)
        if not match:
            raise ValueError("Invalid Python version specified")
        if int(match.group(1)) < 3 or int(match.group(2)) < 6:
            raise ValueError("Otter is only compatible with Python 3.6+")

    template_context = {
        "autograder_dir": ag_config.autograder_dir,
        "otter_env_name": OTTER_ENV_NAME,
        "ottr_version": OTTR_VERSION,
        "channel_priority_strict": channel_priority_strict,
        "has_r_requirements": False,
        "miniforge_version": MINIFORGE_VERSION,
    }

    if plugin_collection is None:
        plugin_collection = PluginCollection(otter_config.get("plugins", []), None, {})

    else:
        plugin_collection.add_new_plugins(otter_config.get("plugins", []))

    plugin_collection.run("during_generate", otter_config, assignment)

    # open requirements if it exists
    extra_requirements, r_requirements = [], None
    with load_default_file(
        requirements,
        lang_config["requirements_filename"],
        default_disabled=no_requirements,
    ) as reqs:
        if reqs is not None:
            if ag_config.lang == "python":
                extra_requirements = [
                    l for l in reqs.split("\n") if l.strip() and not l.strip().startswith("#")
                ]
            elif ag_config.lang == "r":
                r_requirements = reqs
                template_context["has_r_requirements"] = True

    # open environment if it exists
    user_environment = None
    with load_default_file(
        environment, "environment.yml", default_disabled=no_environment
    ) as env_contents:
        template_context["other_environment"] = env_contents
        if env_contents is not None:
            user_environment = yaml.safe_load(env_contents)

    conda_environment = CondaEnvironment(
        python_version or DEFAULT_PYTHON_VERSION,
        ag_config.lang == "r",
        extra_requirements,
        overwrite_requirements,
        user_environment,
        exclude_conda_defaults,
    )

    rendered = {}
    for fn, template in templates.items():
        rendered[fn] = template.render(**template_context)

    if os.path.exists(output_path):
        os.remove(output_path)

    with zipfile.ZipFile(output_path, mode="w") as zf:
        for fn, contents in rendered.items():
            zf.writestr(fn, contents)

        arc_test_dir = "tests"
        pattern = lang_config["test_file_pattern"]
        for file in glob(os.path.join(tests_dir, pattern)):
            zf.write(file, arcname=os.path.join(arc_test_dir, os.path.basename(file)))

        if r_requirements is not None:
            zf.writestr("requirements.r", r_requirements)

        zf.writestr("environment.yml", conda_environment.to_str())

        zf.writestr(OTTER_CONFIG_FILENAME, json.dumps(otter_config, indent=2))

        # copy files into zip file
        if len(files) > 0:
            for file in files:
                full_fp = os.path.abspath(file)
                if os.getcwd() not in full_fp:
                    raise ValueError(f"{file} is not in the working directory")

                relative_fp = pathlib.Path(full_fp).relative_to(pathlib.Path(os.getcwd()))
                if os.path.isfile(full_fp):
                    zf.write(file, arcname=os.path.join("files", relative_fp))
                elif os.path.isdir(full_fp):
                    zip_folder(
                        zf, full_fp, prefix=os.path.join("files", os.path.split(relative_fp)[0])
                    )
                else:
                    raise ValueError(f"Could not find file or directory '{full_fp}'")


if TYPE_CHECKING:
    from ..assign import Assignment



"""Classes for working with test files and test results"""

import json
import math
import nbformat as nbf
import os
import pickle

from typing import Any, Optional, TYPE_CHECKING, TypeVar, Union

from .abstract_test import TestCase, TestCaseResult, TestFile
from .exception_test import ExceptionTestFile, test_case
from .metadata_test import NotebookMetadataExceptionTestFile, NotebookMetadataOKTestFile
from .ok_test import OKTestFile
from .ottr_test import OttrTestFile
from ..nbmeta_config import NBMetadataConfig, OK_FORMAT_VARNAME
from ..utils import format_exception, QuestionNotInLogException


__all__ = [
    "create_test_file",
    "GradingResults",
    "test_case",
    "TestCase",
    "TestFile",
]


T = TypeVar("T")


def create_test_file(
    path: str,
    nbmeta_config: NBMetadataConfig,
    test_name: Optional[str] = None,
) -> TestFile:
    """
    Read a test file or a notebook file and determine the correct ``TestFile`` subclass for this test.

    If ``path`` is not a notebook, the file is executed as a Python script and a global variable is
    used to determine whether the test is OK-formatted or not.

    Args:
        path (``str``): the path to the test file or notebook
        test_name (``str``): the name of the test in the notebook metadata, if ``path`` is
            the path to a Jupyter notebook

    Returns:
        ``TestFile``: an instance of the correct test file

    Raises:
        ``ValueError``: if there is no test name specified and ``path`` is a notebook file
        ``RuntimeError``: if the test file does not define the ``OK_FORMAT`` global variable
    """
    if os.path.splitext(path)[1] == ".ipynb":
        if test_name is None:
            raise ValueError("You must specify a test name when using notebook metadata tests")

        if nbmeta_config.ok_format:
            return NotebookMetadataOKTestFile.from_nbmeta_config(path, nbmeta_config, test_name)

        else:
            return NotebookMetadataExceptionTestFile.from_nbmeta_config(
                path, nbmeta_config, test_name
            )

    env = {}
    with open(path) as f:
        exec(f.read(), env)

    if OK_FORMAT_VARNAME not in env:
        raise RuntimeError(
            f"Malformed test file: does not define the global variable '{OK_FORMAT_VARNAME}'"
        )

    if env[OK_FORMAT_VARNAME]:
        return OKTestFile.from_file(path)

    else:
        return ExceptionTestFile.from_file(path)


class GradingResults:
    """
    Stores and wrangles test result objects

    Initialize with a list of ``otter.test_files.abstract_test.TestFile`` subclass objects and
    this class will store the results as named tuples so that they can be accessed/manipulated easily.
    Also contains methods to put the results into a nice ``dict`` format or into the correct format
    for Gradescope.

    Args:
        results (``list[TestFile]``): the list of test file objects summarized in this grade
    """

    results: dict[str, TestFile]
    """maps test/question names to their ``TestFile`` objects (which store the results)"""

    output: Optional[str]
    """a string to include in the output field for Gradescope"""

    all_hidden: bool
    """whether all results should be hidden from the student on Gradescope"""

    pdf_error: Optional[Exception]
    """
    an error thrown while generating/submitting a PDF of the submission to display to students in
    the Gradescope results
    """

    notebook: Optional[nbf.NotebookNode]
    """the executed notebook with outputs that gave these results"""

    _plugin_data: dict[str, Any]
    """data requested to be stored in the results by plugins"""

    file: Optional[str] = None
    """the submission file that generated these results; not populated by default"""

    _catastrophic_error: Optional[Exception]
    """an error that prevented grading from completing"""

    def __init__(self, test_files: list[TestFile], notebook: Optional[nbf.NotebookNode] = None):
        self.results = {tf.name: tf for tf in test_files}
        self.output = None
        self.all_hidden = False
        self.pdf_error = None
        self.notebook = notebook
        self._catastrophic_error = None
        self._plugin_data = {}

    def __repr__(self):
        return self.summary()

    @classmethod
    def from_ottr_json(cls, ottr_output: dict[str, Any]) -> "GradingResults":
        """
        Creates a ``GradingResults`` object from the JSON output of Ottr (Otter's R client).

        Args:
            ottr_output (``str``): the JSON output of Ottr as a string

        Returns:
            ``GradingResults``: the Ottr grading results
        """
        test_file_results = json.loads(ottr_output)["test_file_results"]
        test_files = []
        for tfr in test_file_results:
            test_cases, test_case_results = [], []

            for tcr in tfr["test_case_results"]:
                tc = tcr["test_case"]
                test_cases.append(
                    TestCase(
                        name=tc["name"],
                        body=tc["code"],
                        hidden=tc["hidden"],
                        points=tc["points"],
                        success_message=tc["success_message"],
                        failure_message=tc["failure_message"],
                    )
                )
                test_case_results.append(
                    TestCaseResult(
                        test_case=test_cases[-1],
                        message=tcr["error"],
                        passed=tcr["passed"],
                    )
                )

            # fix the point values in each test case
            test_cases = TestFile.resolve_test_file_points(tfr.get("points"), test_cases)

            # TestFile.resolve_test_file_points returns a copy of each TestCase, so update the
            # TestCaseResults
            for tc, tcr in zip(test_cases, test_case_results):
                tcr.test_case = tc

            test_file = OttrTestFile(
                name=os.path.splitext(os.path.basename(tfr["filename"]))[0],
                path=tfr["filename"],
                test_cases=test_cases,
                all_or_nothing=False,
            )
            test_file.test_case_results = test_case_results

            test_files.append(test_file)

        return cls(test_files)

    @classmethod
    def without_results(cls, e: Exception) -> "GradingResults":
        """
        Creates an empty results object that represents an execution failure during autograding.

        The returned results object will alert students and instructors to this failure, providing
        the error message and traceback to instructors, and report a score of 0 on Gradescope.

        Args:
            e (``Exception``): the error that was thrown

        Returns:
            ``GradingResults``: the results object
        """
        instc = cls([])
        instc._catastrophic_error = e
        return instc

    @property
    def test_files(self) -> list[str]:
        """the names of all test files tracked in these grading results"""
        return list(self.results.keys())

    @property
    def total(self) -> Union[int, float]:
        """the total points earned"""
        if self._catastrophic_error:
            return 0
        return sum(tr.score for tr in self.results.values())

    @property
    def possible(self) -> Union[int, float]:
        """the total points possible"""
        if self._catastrophic_error:
            return 0
        return sum(tr.possible for tr in self.results.values())

    @property
    def percent(self) -> float:
        """the ratio of points earned, rounded to 4 digits"""
        if self.possible == 0:
            return 0
        return round(self.total / self.possible, 4)

    @property
    def passed_all_public(self) -> bool:
        """whether all public tests in these results passed"""
        return all(tr.passed_all_public for tr in self.results.values())

    @property
    def catastrophic_error(self) -> Optional[Exception]:
        """an error that prevented grading from completing, if applicable"""
        return self._catastrophic_error

    def get_result(self, test_name: str) -> TestFile:
        """
        Returns the ``TestFile`` corresponding to the test with name ``test_name``

        Args:
            test_name (``str``): the name of the desired test

        Returns:
            ``TestFile``: the graded test file object
        """
        return self.results[test_name]

    def get_score(self, test_name: str) -> Union[int, float]:
        """
        Returns the score of a test tracked by these results

        Args:
            test_name (``str``): the name of the test

        Returns:
            ``int | float``: the score
        """
        result = self.results[test_name]
        return result.score

    def update_score(self, test_name: str, new_score: Union[int, float]):
        """
        Override the score for the specified test file.

        Args:
            test_name (``str``): the name of the test file
            new_score (``int | float``): the new score
        """
        self.results[test_name].update_score(new_score)

    def set_output(self, output: str):
        """
        Updates the ``output`` field of the results JSON with text relevant to the entire submission.
        See https://gradescope-autograders.readthedocs.io/en/latest/specs/ for more information.

        Args:
            output (``str``): the output text
        """
        self.output = output

    def clear_results(self):
        """
        Empties the dictionary of results.
        """
        self.results = {}

    def hide_everything(self):
        """
        Indicates that all results should be hidden from students on Gradescope.
        """
        self.all_hidden = True

    def set_plugin_data(self, plugin_name: str, data: Any):
        """
        Stores plugin data for plugin ``plugin_name`` in the results. ``data`` must be picklable.

        Args:
            plugin_name (``str``): the importable name of a plugin
            data (any): the data to store; must be serializable with ``pickle``
        """
        try:
            pickle.dumps(data)
        except:
            raise ValueError(f"Data was not picklable: {data}")
        self._plugin_data[plugin_name] = data

    def get_plugin_data(self, plugin_name: str, default: Optional[T] = None) -> T:
        """
        Retrieves data for plugin ``plugin_name`` in the results.

        This method uses ``dict.get`` to retrieve the data, so a ``KeyError`` is never raised if
        ``plugin_name`` is not found; rather, it returns ``None``.

        Args:
            plugin_name (``str``): the importable name of a plugin
            default (any): a default value to return if ``plugin_name`` is not found

        Returns:
            any: the data stored for ``plugin_name`` if found
        """
        return self._plugin_data.get(plugin_name, default)

    def set_pdf_error(self, error: Exception):
        """
        Set a PDF generation error to be displayed as a failed (0-point) test on Gradescope.

        Args:
            error (``Exception``): the error thrown
        """
        self.pdf_error = error

    def verify_against_log(self, log: "Log", ignore_hidden: bool = True) -> list[str]:
        """
        Verifies these scores against the results stored in this log using the results returned by
        ``Log.get_results`` for comparison. A discrepancy occurs if the scores differ by more than
        the default tolerance of ``math.isclose``. If ``ignore_hidden`` is ``True``, hidden tests
        are ignored when verifying scores.

        Args:
            log (``otter.check.logs.Log``): the log to verify against
            ignore_hidden (``bool``): whether to ignore hidden tests during verification

        Returns:
            ``list[str]``: a list of error messages for discrepancies; if none were found, the list
                is empty
        """
        l = []
        for test_name, test_file in self.results.items():
            if ignore_hidden:
                tcrs = [
                    test_case_result
                    for test_case_result in test_file.test_case_results
                    if not test_case_result.test_case.hidden
                ]
                score = sum(tcr.test_case.points for tcr in tcrs if tcr.passed)
            else:
                score = test_file.score
            try:
                tf = log.get_results(test_name)
                if not isinstance(tf, TestFile):
                    l.append(f"No score logged for {test_name}")
                    continue
                if not math.isclose(score, tf.score):
                    l.append(
                        f"Score for {test_name} ({score:.3f}) differs from logged score "
                        f"({tf.score:.3f})"
                    )
            except QuestionNotInLogException:
                l.append(f"No score for {test_name} found in this log")
        return l

    def to_report_str(self) -> str:
        """
        Returns these results as a report string generated using the ``__repr__`` of the
        ``TestFile`` class.

        Returns:
            ``str``: the report
        """
        return "\n".join(repr(test_file) for test_file in self.test_files)

    def to_dict(self) -> dict[str, Any]:
        """
        Converts these results into a dictionary, extending the fields of the named tuples in
        ``results`` into key, value pairs in a ``dict``.

        Returns:
            ``dict``: the results in dictionary form
        """
        return {tn: tf.to_dict() for tn, tf in self.results.items()}

    def summary(self, public_only: bool = False) -> str:
        """
        Generate a summary of these results and return it as a string.

        Args:
            public_only (``bool``): whether only public test cases should be included

        Returns:
            ``str``: the summary of results
        """
        if self._catastrophic_error:
            return str(self._catastrophic_error)
        return "\n\n".join(tf.summary(public_only=public_only) for _, tf in self.results.items())

    def has_catastrophic_failure(self) -> bool:
        """
        Returns whether these results contain a catastrophic error (i.e. an error that prevented
        submission results from being generated or read).

        Returns:
            ``bool``: whether there is such an error
        """
        return self._catastrophic_error is not None

    def to_gradescope_dict(self, ag_config: "AutograderConfig") -> dict[str, Any]:
        """
        Convert these results into a dictionary formatted for Gradescope's autograder.

        Args:
            ag_config (``otter.run.run_autograder.autograder_config.AutograderConfig``): the
                autograder config

        Returns:
            ``dict``: the results formatted for Gradescope
        """
        output = {"tests": []}

        if self._catastrophic_error:
            output["tests"].append(
                {
                    "name": "Autograder Failed",
                    "visibility": "visible",
                    "output": "The autograder failed to produce any results. Please alert your instructor to this failure for assistance in debugging it.",
                    "status": "failed",
                }
            )
            tb = format_exception(self._catastrophic_error)
            output["tests"].append(
                {
                    "name": "Autograder Exception",
                    "visibility": "hidden",
                    "output": f"The exception below was thrown when attempting to read the results from executing the notebook. (This message is not visible to students.)\n\n{tb}",
                    "status": "failed",
                }
            )
            output["score"] = 0
            return output

        if self.output is not None:
            output["output"] = self.output

        # hidden visibility determined by show_hidden
        hidden_test_visibility = ("hidden", "after_published")[ag_config.show_hidden]

        # if show_all_public is true and all tests are public tests, display all tests in results
        if ag_config.show_all_public and all(tf.all_public for tf in self.results.values()):
            hidden_test_visibility = "visible"

        # start w/ summary of public tests
        if not ag_config.show_hidden or ag_config.force_public_test_summary:
            output["tests"].append(
                {
                    "name": "Public Tests",
                    "visibility": "visible",
                    "output": self.summary(public_only=True),
                    "status": "passed" if self.passed_all_public else "failed",
                }
            )

        # add PDF error test if indicated
        if ag_config.warn_missing_pdf and self.pdf_error is not None:
            output["tests"].append(
                {
                    "name": "PDF Generation Failed",
                    "visibility": "visible",
                    "output": str(self.pdf_error),
                    "status": "failed",
                }
            )

        for test_name in self.test_files:
            test_file = self.get_result(test_name)
            score, possible = test_file.score, test_file.possible

            output["tests"].append(
                {
                    "name": test_file.name,
                    "score": round(score, 5),
                    "max_score": round(possible, 5),
                    "visibility": hidden_test_visibility,
                    "output": test_file.summary(),
                }
            )

        if ag_config.show_stdout:
            output["stdout_visibility"] = "after_published"

        if ag_config.points_possible is not None:
            try:
                output["score"] = self.total / self.possible * ag_config.points_possible
            except ZeroDivisionError:
                output["score"] = 0

        if ag_config.score_threshold is not None:
            try:
                if self.total / self.possible >= ag_config.score_threshold:
                    output["score"] = ag_config.points_possible or self.possible
                else:
                    output["score"] = 0
            except ZeroDivisionError:
                if 0 >= ag_config.score_threshold:
                    output["score"] = ag_config.points_possible or self.possible
                else:
                    output["score"] = 0

        if self.all_hidden:
            for test in output["tests"]:
                test["visibility"] = "hidden"
            output["stdout_visibility"] = "hidden"

        return output


if TYPE_CHECKING:
    from ..check.logs import Log
    from ..run import AutograderConfig



"""Runner classes for different languages"""

from typing import Any

from ..autograder_config import AutograderConfig


def create_runner(otter_config: dict[str, Any], **kwargs: Any):
    """
    Return an instantiated runner for the assignment based on user-specified configurations.
    """
    config = AutograderConfig({**otter_config, **kwargs})

    if config.lang == "python":
        from .python_runner import PythonRunner

        return PythonRunner(config)
    elif config.lang == "r":
        from .r_runner import RRunner

        return RRunner(config)
    # There is no else required here because AutograderConfig validates that lang is one of the set
    # of possible values enumerated in the conditions above.



"""Assignment block parsing for Otter Assign"""

import nbformat
import re
import yaml

from enum import Enum
from typing import Any

from .utils import is_cell_type
from ..utils import get_source


class BlockType(Enum):
    """
    An enum of allowed block types.
    """

    QUESTION = "question"
    PROMPT = "prompt"
    SOLUTION = "solution"
    TESTS = "tests"


def extract_fenced_otter_cell(cell: nbformat.NotebookNode) -> nbformat.NotebookNode:
    """
    Converts a Markdown config cell to a raw config cell.

    A Markdown cell with the contents

    .. code-block::

        ```otter
        <cell contents>
        ```

    would become a raw cell with the contents

    .. code-block::

        <cell contents>

    If the contents are not wrapped in a code block with the language set to ``otter``, the cell is
    returned unchanged.

    Args:
        cell (``nbformat.NotebookNode``): the cell to extract the config from

    Returns:
        ``nbformat.NotebookNode``: the unfenced cell contents
    """
    if not is_cell_type(cell, "markdown"):
        return cell

    source = get_source(cell)
    if (
        source[0].strip() == "```otter"
        and all(not l.strip() == "```" for l in source[1:-1])
        and source[-1].strip() == "```"
    ):
        return nbformat.v4.new_raw_cell("\n".join(source[1:-1]))

    return cell


def is_block_boundary_cell(
    cell: nbformat.NotebookNode,
    block_type: BlockType,
    end: bool = False,
) -> bool:
    """
    Determine whether ``cell`` is a boundary cell for a ``block_type`` block. If ``end`` is true,
    the block should be an end block; otherwise, it should be a begin block.

    Args:
        cell (``nbformat.NotebookNode``): the cell to check
        block_type (``BlockType``): the block type to check for
        end (``bool``): whether to check for an end boundary instead of a begin

    Returns:
        ``bool``: whether the cell is a boundary cell of type ``block_type``
    """
    cell = extract_fenced_otter_cell(cell)
    begin_or_end = "end" if end else "begin"
    regex = rf"#\s+{ begin_or_end }\s+{ block_type.value }\s*"
    source = get_source(cell)
    return is_cell_type(cell, "raw") and bool(re.match(regex, source[0], flags=re.IGNORECASE))


def is_assignment_config_cell(cell: nbformat.NotebookNode) -> bool:
    """
    Determine whether ``cell`` is an assignment configuration cell.

    An assignment configuration cell is a raw cell starting with the line ``# ASSIGNMENT CONFIG``,
    e.g.

    .. code-block:: yaml

        # ASSIGNMENT CONFIG
        requirements: requirements.txt
        files:
            - data.csv

    Args:
        cell (``nbformat.NotebookNode``): the cell to check

    Returns:
        ``bool``: whether the cell is an assignment config cell
    """
    cell = extract_fenced_otter_cell(cell)
    regex = r"#\s+assignment\s+config\s*"
    source = get_source(cell)
    return is_cell_type(cell, "raw") and bool(re.match(regex, source[0], flags=re.IGNORECASE))


def get_cell_config(cell: nbformat.NotebookNode) -> dict[str, Any]:
    """
    Parse a cell's contents as YAML and return the resulting dictionary.

    Args:
        cell (``nbformat.NotebookNode``): the cell to check

    Returns:
        ``dict[str, object]``: the parsed configurations

    Raises:
        ``TypeError``: if parsing the YAML does not return a dictionary
    """
    source = get_source(extract_fenced_otter_cell(cell))
    config = yaml.full_load("\n".join(source))
    if config is None:
        config = {}
    if not isinstance(config, dict):
        raise TypeError(f"Found a begin cell configuration that is not a dictionary: {cell}")
    return config



"""Utilities for Otter Generate"""

import os
import zipfile

from typing import Any, Optional, Union


def zip_folder(
    zf: zipfile.ZipFile, path: str, prefix: str = "", exclude: Optional[list[str]] = None
):
    """
    Recursively add the contents of a directory into a ``zipfile.ZipFile``.

    Args:
        zf (``zipfile.ZipFile``): the open zip file object to add to
        path (``str``): an absolute path to the directory to add to the zip file
        prefix (``str``): a prefix to add to the basename of the path for the archive name
    """
    if exclude is None:
        exclude = []

    if not os.path.isabs(path):
        raise ValueError("'path' must be absolute path")

    parent_basename = os.path.basename(path)
    for file in os.listdir(path):
        if file in exclude:
            continue
        child_path = os.path.join(path, file)

        if os.path.isfile(child_path):
            arcname = os.path.join(prefix, parent_basename, file)
            zf.write(child_path, arcname=arcname)

        elif os.path.isdir(child_path):
            zip_folder(zf, child_path, prefix=os.path.join(prefix, parent_basename))


def _get_dep_name(d: str) -> str:
    """
    Determine the name of a dependency, ignoring version information.

    Args:
        d (``str``): the dependency

    Returns:
        ``str``: the dependency name
    """
    if "+" in d:
        return d
    return d.split(">")[0].split("<")[0].split("=")[0]


def merge_conda_environments(e1: dict[str, Any], e2: dict[str, Any], name: str) -> dict[str, Any]:
    """
    Merge two conda environments into a single environment with the specified name.

    Merges the channels, dependencies, and pip dependencies of two conda environments. If ``e1`` and
    ``e2`` conflict, the values in ``e1`` are taken.

    Args:
        e1 (``dict[str, Any]``): the first conda environment
        e2 (``dict[str, Any]``): the second conda environment
        name (``str``): the resulting environment name

    Returns:
        ``dict[str, Any]``: the merged conda environment
    """
    e: dict[str, Any] = {"name": name, "channels": e1.get("channels", []), "dependencies": []}
    e["channels"].extend([c for c in e2.get("channels", []) if c not in e["channels"]])

    if not e["channels"]:
        e.pop("channels")

    seen_deps, dicts = set(), []

    def handle_dep(d: Union[str, dict[str, Any]], target: list[str]):
        if isinstance(d, dict):
            dicts.append(d)
            return
        if not isinstance(d, str):
            raise TypeError(f"Value of invalid type in dependencies list: {d}")
        dep_name = _get_dep_name(d)
        if dep_name not in seen_deps:
            seen_deps.add(dep_name)
            target.append(d)

    for ei in [e1, e2]:
        for d in ei["dependencies"]:
            handle_dep(d, e["dependencies"])

    if len(dicts) > 2:
        raise ValueError("Too many dictionaries found in environment dependencies")

    if len(dicts) == 1:
        e["dependencies"].append(dicts[0])

    elif len(dicts) == 2:
        target = []
        e["dependencies"].append({"pip": target})
        for di in dicts:
            if set(di.keys()) != {"pip"}:
                raise ValueError("dictionaries should only contain the pip key")
            for d in di["pip"]:
                handle_dep(d, target)

    return e



"""PDF via HTML exporter"""

import os

from typing import Any

from .base_exporter import BaseExporter, TEMPLATE_DIR


# nbconvert can't be used on jupyterlite, so try importing it and raise an error if it's not found
# when the exporter is used
_NBCONVERT_ERROR = None
try:
    import nbconvert
except ImportError as e:
    nbconvert = None
    _NBCONVERT_ERROR = e


class PDFViaHTMLExporter(BaseExporter):
    """
    An exporter that uses nbconvert's WebPDF exporter to convert notebooks to PDFs via HTML.

    Attributes:
        default_options (``dict``): the default options for this exporter
    """

    default_options = BaseExporter.default_options.copy()
    default_options.update({"save_html": False, "template": "via_html"})

    @classmethod
    def convert_notebook(cls, nb_path: str, dest: str, **kwargs: Any):
        if nbconvert is None:
            raise _NBCONVERT_ERROR

        options = cls.default_options.copy()
        options.update(kwargs)

        nb = cls.load_notebook(
            nb_path, filtering=options["filtering"], pagebreaks=options["pagebreaks"]
        )

        nbconvert.TemplateExporter.extra_template_basedirs = [str(TEMPLATE_DIR)]
        orig_template_name = nbconvert.TemplateExporter.template_name
        nbconvert.TemplateExporter.template_name = options["template"]

        exporter = nbconvert.WebPDFExporter()

        try:
            pdf, _ = nbconvert.export(exporter, nb)
        except RuntimeError as e:
            # Replace nbconvert's error about installing chromium since their flag can't be passed
            # to Otter.
            if "--allow-chromium-download" in str(e):
                raise RuntimeError(
                    "No suitable version of chromium was found; please install chromium by running "
                    "'playwright install chromium'"
                )
            raise e

        pdf_path = os.path.splitext(dest)[0] + ".pdf"
        with open(pdf_path, "wb+") as f:
            f.write(pdf)

        nbconvert.TemplateExporter.template_name = orig_template_name



"""A programmatic API for using Otter-Grader"""

__all__ = ["export_notebook", "grade_submission"]

import os

from contextlib import nullcontext, redirect_stdout
from typing import Optional

from .export import export_notebook
from .run import main as run_grader
from .test_files import GradingResults


def grade_submission(
    submission_path: str,
    ag_path: str = "autograder.zip",
    *,
    quiet: bool = False,
    debug: bool = False,
    extra_submission_files: Optional[list[str]] = None,
) -> GradingResults:
    """
    Runs non-containerized grading on a single submission at ``submission_path`` using the autograder
    configuration file at ``ag_path``.

    Creates a temporary grading directory using the ``tempfile`` library and grades the submission
    by replicating the autograder tree structure in that folder and running the autograder there. Does
    not run environment setup files (e.g. ``setup.sh``) or install requirements, so any requirements
    should be available in the environment being used for grading.

    Print statements executed during grading can be suppressed with ``quiet``.

    Args:
        submission_path (``str``): path to submission file
        ag_path (``str``): path to autograder zip file
        quiet (``bool``): whether to suppress print statements during grading; default
            ``False``
        debug (``bool``): whether to run the submission in debug mode (without ignoring
            errors)
        extra_submission_files (``list[str] | None``): extra files to copy into the submission
            directory; this should really only be used internally by Otter, so use at your own risk

    Returns:
        ``otter.test_files.GradingResults``: the results object produced during the grading of the
            submission.
    """
    if quiet:
        f = open(os.devnull, "w")
        cm = redirect_stdout(f)
    else:
        cm = nullcontext()

    with cm:
        results = run_grader(
            submission_path,
            autograder=ag_path,
            output_dir=None,
            no_logo=True,
            debug=debug,
            extra_submission_files=extra_submission_files,
        )

    if quiet:
        f.close()

    return results



"""Utilities for Otter Grade"""

import os
import pandas as pd
import re

from python_on_whales import docker

from ..test_files import GradingResults


OTTER_DOCKER_IMAGE_NAME = "otter-grade"

POINTS_POSSIBLE_LABEL = "points-per-question"

SCORES_DICT_FILE_KEY = "file"
SCORES_DICT_GRADING_STATUS_KEY = "grading_status"
SCORES_DICT_PERCENT_CORRECT_KEY = "percent_correct"
SCORES_DICT_TOTAL_POINTS_KEY = "total_points_earned"


class TimeoutException(Exception):
    """
    This Exception is thrown when grading a notebook exceeds the timeout value specified.
    """

    pass


def list_files(path: str) -> list[str]:
    """
    Returns a list of all non-hidden files in a directory

    Args:
        path (``str``): path to a directory

    Returns:
        ``list[str]``: list of filenames in the given directory

    """
    return [
        file
        for file in os.listdir(path)
        if os.path.isfile(os.path.join(path, file)) and file[0] != "."
    ]


def merge_csv(dataframes: list[pd.DataFrame]) -> pd.DataFrame:
    """
    Merges dataframes along the vertical axis

    Args:
        dataframes (``list[pandas.core.frame.DataFrame]``): list of dataframes with same columns

    Returns:
        ``pandas.core.frame.DataFrame``: a merged dataframe resulting from 'stacking' all input
            dataframes

    """
    final_dataframe = pd.concat(dataframes, axis=0, join="inner")
    do_not_sort = final_dataframe[final_dataframe["file"] == POINTS_POSSIBLE_LABEL]
    sort_these = final_dataframe[final_dataframe["file"] != POINTS_POSSIBLE_LABEL]
    df_sorted = sort_these.sort_values(by="file")
    df_final = pd.concat([do_not_sort, df_sorted], ignore_index=True)
    return df_final


def prune_images(force: bool = False):
    """
    Prunes all Docker images named ``otter-grade``.
    """
    images = docker.images(OTTER_DOCKER_IMAGE_NAME)
    print("The following images will be deleted:")
    for image in images:
        print(f"    {image.repo_tags[0]}")

    if not force:
        sure = input(
            "Are you sure you want to delete these images? This action cannot be undone [y/N] "
        )
        sure = bool(re.match(r"ye?s?", sure, flags=re.IGNORECASE))

    else:
        sure = True

    if sure:
        for image in images:
            image.remove(force=True)

        print("Images deleted.")

    else:
        print("Prune cancelled.")


def get_points_possible_df(scores: list[GradingResults]) -> pd.DataFrame:
    """
    From the list of ``GradingResults`` objects try to find one that does not have a catastrophic
    failure and extract the points-per-question into its own DataFrame. This becomes the
    first row of final_grades.csv. If all the ``GradingResults`` objects  are catastrophic failures
    then returns an empty DataFrame

    Args:
        scores (``list[otter.test_files.GradingResults]``): the score objects to locate the points
            per question

    Returns:
        ``pd.DataFrame``: the points per question dataframe; if all failures empty dataframe
    """
    gr_completed = [s for s in scores if not s.has_catastrophic_failure()]
    pts_poss_df = pd.DataFrame()
    if gr_completed:
        pts_poss_dict = {t: [d["possible"]] for t, d in gr_completed[0].to_dict().items()}
        pts_poss_dict[SCORES_DICT_FILE_KEY] = [POINTS_POSSIBLE_LABEL]
        pts_poss_dict[SCORES_DICT_PERCENT_CORRECT_KEY] = [1]
        pts_poss_dict[SCORES_DICT_TOTAL_POINTS_KEY] = [gr_completed[0].possible]
        pts_poss_dict[SCORES_DICT_GRADING_STATUS_KEY] = ["--"]
        pts_poss_df = pd.DataFrame(pts_poss_dict)

    return pts_poss_df


def merge_scores_to_df(scores: list[GradingResults]) -> pd.DataFrame:
    """
    Convert a list of ``GradingResults`` objects to a scores dataframe, including a row
    with the total point values for each question.

    Args:
        scores (``list[otter.test_files.GradingResults]``): the score objects to merge

    Returns:
        ``pd.DataFrame``: the scores dataframe
    """
    dfs = []
    pts_poss_df = get_points_possible_df(scores)
    if not pts_poss_df.empty:
        dfs.append(pts_poss_df)

    for gr in scores:
        scores_dict = gr.to_dict()
        failed = gr.has_catastrophic_failure()
        if failed and not pts_poss_df.empty:
            scores_dict = {t: [0] for t in pts_poss_df.to_dict()}
        elif not failed:
            scores_dict = {t: [scores_dict[t]["score"]] for t in scores_dict}

        grading_status = "Completed" if not failed else str(gr.catastrophic_error)

        scores_dict[SCORES_DICT_TOTAL_POINTS_KEY] = gr.total
        scores_dict[SCORES_DICT_FILE_KEY] = gr.file
        scores_dict[SCORES_DICT_PERCENT_CORRECT_KEY] = gr.percent
        scores_dict[SCORES_DICT_GRADING_STATUS_KEY] = grading_status

        dfs.append(pd.DataFrame(scores_dict, index=[0]))

    output_df = merge_csv(dfs)
    cols = output_df.columns.tolist()
    question_cols = sorted(
        c
        for c in cols
        if c
        not in {
            SCORES_DICT_FILE_KEY,
            SCORES_DICT_TOTAL_POINTS_KEY,
            SCORES_DICT_PERCENT_CORRECT_KEY,
            SCORES_DICT_GRADING_STATUS_KEY,
        }
    )
    output_df = output_df[
        [
            SCORES_DICT_FILE_KEY,
            *question_cols,
            SCORES_DICT_TOTAL_POINTS_KEY,
            SCORES_DICT_PERCENT_CORRECT_KEY,
            SCORES_DICT_GRADING_STATUS_KEY,
        ]
    ]

    return output_df



"""Assignment tests manager for Otter Assign"""

import ast
import nbformat as nbf
import os
import pandas as pd
import pathlib
import pprint
import re
import yaml

from jinja2 import Template
from typing import Any, Union

from .assignment import Assignment
from .question_config import QuestionConfig
from .utils import str_to_doctest
from ..nbmeta_config import NBMetadataConfig, OK_FORMAT_VARNAME
from ..test_files.abstract_test import TestCase, TestFile
from ..utils import get_source


BEGIN_TEST_CONFIG_REGEX = r'(?:.\s*=\s*)?""?"?\s*#\s*BEGIN\s*TEST\s*CONFIG'
END_TEST_CONFIG_REGEX = r'""?"?\s*#\s*END\s*TEST\s*CONFIG'

EXCEPTION_BASED_TEST_FILE_TEMPLATE = Template(
    """\
from otter.test_files import test_case

{{ OK_FORMAT_VARNAME }} = False

name = "{{ name }}"
points = {{ points }}{% if all_or_nothing %}
all_or_nothing = True{% endif %}

{% for tc in test_cases %}@test_case(points={{ tc.points }}, hidden={{ tc.hidden }}{% if tc.success_message is not none %}, 
    success_message="{{ tc.success_message }}"{% endif %}{% if tc.failure_message is not none %}, 
    failure_message="{{ tc.failure_message }}"{% endif %})
{{ tc.body }}

{% endfor %}
"""
)


class AssignmentTestsManager:
    """
    A class for creating and managing test cases for an assignment.

    Args:
        assignment (``otter.assign.assignment.Assignment``): the assignment config
    """

    assignment: Assignment
    """the assignment config"""

    _tests_by_question: dict[str, list[TestCase]]
    """a dictionary mapping question names to lists of test cases"""

    _questions: dict[str, QuestionConfig]
    """a dictionary mapping question names to ``QuestionConfig`` objects"""

    def __init__(self, assignment: Assignment):
        self.assignment = assignment
        self._tests_by_question = {}
        self._questions = {}

    def any_public_tests(self, question: QuestionConfig) -> bool:
        """
        Determine whether any of the test cases in the specified question are public.

        Args:
            question (``otter.assign.question_config.QuestionConfig``): the question config

        Returns:
            ``bool``: whether any of the test cases for the question are public
        """
        return any(not tc.hidden for tc in self._tests_by_question[question["name"]])

    def _add_test_case(self, question: QuestionConfig, test_case: TestCase):
        """
        Track a test case for the specified question.

        Args:
            question (``otter.assign.question_config.QuestionConfig``): the question config
            test_case (``TestCase``): the test case to track
        """
        self.add_question(question)
        self._tests_by_question[question.name].append(test_case)

    def _parse_test_config(self, source: list[str]) -> tuple[dict[str, Any], Union[int, None]]:
        """
        Parse test configurations from the test cell source.

        Args:
            source (``list[str]``): the source lines of the test cell

        Returns:
            ``tuple[dict, int | None]``: the parsed config and the index in ``source`` of the line
                at which the config ended, if any
        """
        config, i = {}, None
        if re.match(BEGIN_TEST_CONFIG_REGEX, source[0], flags=re.IGNORECASE):
            for i, line in enumerate(source):
                if re.match(END_TEST_CONFIG_REGEX, line, flags=re.IGNORECASE):
                    break
            config = yaml.full_load("\n".join(source[1:i]))
            assert isinstance(config, dict), f"Invalid test config in cell {source}"

        return config, i

    def add_question(self, question: QuestionConfig):
        """
        Add a question for tracking and inclusion in the assignment summary

        Args:
            question (``otter.assign.question_config.QuestionConfig``): the question config
        """
        self._questions[question.name] = question
        if question.name not in self._tests_by_question:
            self._tests_by_question[question.name] = []

    def read_test(self, cell: nbf.NotebookNode, question: QuestionConfig):
        """
        Parse and track a test case from the provided cell for the specified question.

        Args:
            cell (``nbformat.NotebookNode``): a test cell
            question (``otter.assign.question_config.QuestionConfig``): the question config
        """
        source = get_source(cell)
        ensure_valid_syntax(source, question)

        if source[0].lstrip().startswith("#"):
            hidden = bool(re.search(r"\bhidden\b", source[0], flags=re.IGNORECASE))
        else:
            hidden = False

        test_start_line = 0 if hidden else -1

        output = ""
        for o in cell["outputs"]:
            output += "".join(o.get("text", ""))
            results = o.get("data", {}).get("text/plain")
            if results and isinstance(results, list):
                output += results[0]
            elif results:
                output += results

        config, maybe_test_start_line = self._parse_test_config(source)
        test_start_line = (
            maybe_test_start_line if maybe_test_start_line is not None else test_start_line
        )

        hidden = config.get("hidden", hidden)
        points = config.get("points", None)
        success_message = config.get("success_message", None)
        failure_message = config.get("failure_message", None)

        test_source = get_source(cell)[test_start_line + 1 :]
        test_ast = ast.parse("\n".join(test_source))
        body: str = ""
        if self.assignment.tests.ok_format:
            inp = ast.unparse(_AnnotationRemover().visit(test_ast))
            body_lines = str_to_doctest(inp.split("\n"), [])
            body_lines.extend(output.split("\n"))
            body = "\n".join(body_lines)
        else:
            if len(test_ast.body) > 2 or not isinstance(test_ast.body[0], ast.FunctionDef):
                raise ValueError(
                    f"Error in question {question.name}: an exception-based test cell may have at "
                    "most a function definition and a call to that function definition but a test "
                    f"cell with {len(test_ast.body)} nodes in its AST was encountered"
                )
            body = ast.unparse(test_ast.body[0])

        self._add_test_case(
            question,
            # TODO: does the test case need a name?
            TestCase("", body, hidden, points, success_message, failure_message),
        )

    def has_tests(self, question: QuestionConfig) -> bool:
        """
        Determine whether the specified question has any test cases.

        Args:
            question (``otter.assign.question_config.QuestionConfig``): the question config

        Returns:
            ``bool``: whether the question has any test cases
        """
        return len(self._tests_by_question.get(question.name, [])) > 0

    @staticmethod
    def _resolve_test_file_points(
        total_points: Union[int, float, list[Union[int, float, None]], None],
        test_cases: list[TestCase],
    ) -> list[TestCase]:
        """
        Validate and reformat the point values of the provided test cases taking into account the
        total points for the question.

        For Python, this is validation doesn't change the underlying test cases and only raises an
        error if the specified point values are invalid. This may not be the case for other
        languages, however.

        Args:
            total_points (``int | float | list[int | float | None] | None``): the value of the
                ``points`` configuration for the question
            test_cases (``list[TestCase]``): the test cases for the question

        Returns:
            ``list[TestCase]``: updated test cases
        """
        TestFile.resolve_test_file_points(total_points, test_cases)
        return test_cases

    def _create_test_file_info(self, question_name: str) -> dict[str, Any]:
        """
        Create a ``dict`` containing the test file information for the question with the specified
        name.

        Args:
            question_name (``str``): the question name

        Returns:
            ``dict``: the test file info
        """
        question = self._questions[question_name]
        test_cases = self._tests_by_question[question_name]

        points = question.points
        if isinstance(points, dict):
            points = points.get("each", 1) * len(test_cases)
        elif isinstance(points, list):
            if len(points) != len(test_cases):
                raise ValueError(
                    f"Error in question {question.name}: length of 'points' is {len(points)} "
                    f"but there are {len(test_cases)} tests"
                )

        # check for errors in resolving points
        test_cases = self._resolve_test_file_points(points, test_cases)

        return {
            "name": question.name,
            "points": points,
            "all_or_nothing": question.all_or_nothing,
            "test_cases": test_cases,
        }

    @staticmethod
    def _create_ok_test_case(test_case: TestCase) -> dict[str, Any]:
        """
        Create an OK-formatted test case for a test case object.

        Args:
            test (``TestCase``): the test case to convert to OK format

        Returns:
            ``dict``: the OK-formatted test case
        """
        ret = {
            "code": test_case.body,
            "hidden": test_case.hidden,
            "locked": False,
        }

        if test_case.points is not None:
            ret["points"] = test_case.points
        if test_case.success_message:
            ret["success_message"] = test_case.success_message
        if test_case.failure_message:
            ret["failure_message"] = test_case.failure_message

        return ret

    @classmethod
    def _create_ok_test_suite(cls, test_cases: list[TestCase]) -> dict[str, Any]:
        """
        Create an OK-formatted test suite for a list of test cases.

        Args:
            test_cases (``list[TestCase]``): the test cases

        Returns:
            ``dict``: the OK-formatted test suite
        """
        return {
            "cases": [cls._create_ok_test_case(tc) for tc in test_cases],
            "scored": True,
            "setup": "",
            "teardown": "",
            "type": "doctest",
        }

    def _format_test(
        self,
        name: str,
        points: Union[int, float, list[Union[int, float, None]], None],
        all_or_nothing: bool,
        test_cases: list[TestCase],
    ) -> Union[str, dict[str, Any]]:
        """
        Format the test cases for a question based on the assignment config.

        Args:
            name (``str``): the name of the question
            points (``int | float | list[int | float | None] | None``): the ``points`` configuration
                from the question config
            all_or_nothing (``bool``): the ``all_or_nothing`` configuration from the question
                config
            test_cases (``list[TestCase]``): the test cases for the question

        Returns:
            ``dict | str``: the formatted test file
        """
        if self.assignment.tests.ok_format:
            test = {
                "name": name,
                "points": points,
            }
            # Only set all_or_nothing if it's true, since it defaults to false.
            if all_or_nothing:
                test["all_or_nothing"] = True
            test["suites"] = [self._create_ok_test_suite(test_cases)]

        else:
            template_kwargs = {
                "name": name,
                "points": points,
                "all_or_nothing": all_or_nothing,
                "test_cases": test_cases,
                "OK_FORMAT_VARNAME": OK_FORMAT_VARNAME,
            }
            test = EXCEPTION_BASED_TEST_FILE_TEMPLATE.render(**template_kwargs)

        return test

    def write_tests(
        self,
        nbmeta_config: NBMetadataConfig,
        test_dir: Union[pathlib.Path, str],
        include_hidden: bool = True,
        force_files: bool = False,
    ):
        """
        Write all test files to a notebook's metadata or a tests directory.

        Args:
            nbmeta_config (``otter.nbmeta_config.NBMetadataConfig``): the notebook metadata config
            test_dir (``pathlib.Path | str``): the tests directory
            include_hidden (``bool``): whether to include hidden test cases
            force_files (``bool``): whether to force writing to test files (overriding the
                assignment config)
        """
        use_files = self.assignment.tests.files or force_files
        if not use_files:
            nbmeta_config.tests = {}
            nbmeta_config.ok_format = self.assignment.tests.ok_format

        test_ext = ".py" if self.assignment.is_python else ".R"
        for test_name in self._tests_by_question.keys():
            if not self._tests_by_question[test_name]:
                continue

            test_info = self._create_test_file_info(test_name)
            test_path = os.path.join(test_dir, test_name + test_ext)

            if not include_hidden:
                test_info["test_cases"] = [tc for tc in test_info["test_cases"] if not tc.hidden]
                if isinstance(test_info["points"], list):
                    test_info["points"] = [
                        p
                        for tc, p in zip(test_info["test_cases"], test_info["points"])
                        if not tc.hidden
                    ]

            test = self._format_test(
                test_info["name"],
                test_info["points"],
                test_info["all_or_nothing"],
                test_info["test_cases"],
            )

            if use_files:
                with open(test_path, "w+") as f:
                    if isinstance(test, dict):
                        f.write(f"{OK_FORMAT_VARNAME} = True\n\ntest = ")
                        pprint.pprint(test, f, indent=4, width=200, depth=None)

                    else:
                        f.write(test)

            else:
                nbmeta_config.tests[test_info["name"]] = test

    def determine_question_point_value(self, question: QuestionConfig) -> Union[int, float]:
        """
        Determine the point value of a question using the question config and its test cases.

        Args:
            question (``otter.assign.question_config.QuestionConfig``): the question config

        Returns:
            ``int | float``: the point value of the question
        """
        if question.manual:
            return question.points or 0

        test_cases = self._tests_by_question.get(question.name, [])

        points = question.points
        if isinstance(points, dict):
            points = points.get("each", 1) * len(test_cases)

        if len(test_cases) == 0:
            if points is None and question.manual:
                raise ValueError(
                    f"Point value unspecified for question with no test cases: {question.name}"
                )

            return points if points is not None else 1

        try:
            resolved_test_cases = TestFile.resolve_test_file_points(points, test_cases)
        except Exception as e:
            raise type(e)(f'Error in "{question.name}" test cases: {e}')

        points = round(sum(tc.points for tc in resolved_test_cases), 5)
        return int(points) if points % 1 == 0 else points

    def generate_assignment_summary(self) -> str:
        """
        Generate a summary of the assignment's questions.

        Returns:
            ``str``: the summary
        """
        rows, manual, autograded, total = [], 0, 0, 0
        for question_name in sorted(self._tests_by_question.keys()):
            config = self._questions[question_name]
            points = self.determine_question_point_value(config)
            rows.append({"name": question_name, "points": points, "manual": config.manual})
            total += points
            if config.manual:
                manual += points
            else:
                autograded += points

        summary = f"Assignment summary:\n"
        summary += f"Total points: {total}\n"
        summary += f"Autograded:   {autograded}\n"
        summary += f"Manual:       {manual}\n\n"
        summary += str(pd.DataFrame(rows))
        return summary


def ensure_valid_syntax(source: list[str], question: QuestionConfig) -> None:
    """
    Check that a cell's source is valid Python syntax.

    Args:
        source (``list[str]``): the cell source lines

    Raises:
        ``ValueError``: if invalid syntax is found
    """
    source = "\n".join(source)
    try:
        ast.parse(source)
    except SyntaxError:
        raise ValueError(
            f"A test cell in question {question.name} contains invalid Python syntax:\n{source}"
        )


class _AnnotationRemover(ast.NodeTransformer):
    """
    An AST node transformer that removes all type annotations.
    """

    def visit(self, node: Any):
        self.generic_visit(node)
        if hasattr(node, "annotation"):
            node.annotation = None
        if hasattr(node, "returns"):
            node.returns = None
        if hasattr(node, "type_params"):
            node.type_params = None
        return node



"""
Abstract base plugin class for Otter
"""

from abc import ABC


class PluginEventNotSupportedException(Exception):
    """
    Exception to raise when an event is not supported by a plugin.
    """


class AbstractOtterPlugin(ABC):
    """
    Abstract base class for Otter plugins to inherit from. Includes the following methods:

    - ``during_assign``: run during Otter Assign after output directories are written
    - ``during_generate``: run during Otter Generate while all files are in-memory and before the
      the `tmp` directory is created
    - ``from_notebook``: run as students as they work through the notebook; see ``Notebook.run_plugin``
    - ``notebook_export``: run by ``Notebook.export`` for adding files to the export zip file
    - ``before_grading``: run before the submission is executed for altering configurations
    - ``before_execution``: run before the submission is executed for altering the submission
    - ``after_grading``: run after all tests are run and scores are assigned
    - ``generate_report``: run after results are written

    See the docstring for the default implementation of each method for more information, including
    arguments and return values. If plugin has no actions for any event above, that method should raise
    ``otter.plugins.PluginEventNotSupportedException``. (This is the default behavior of this ABC, so
    inheriting from this class will do this for you for any methods you don't overwrite.)

    If this plugin requires metadata, it should be included in the plugin configuration of the
    ``otter_config.json`` file as a subdictionary with key a key corresponding to the importable name
    of the plugin. If no configurations are required, the plugin name should be listed as a string.
    For example, the config below provides configurations for ``MyOtterPlugin`` but not
    ``MyOtherOtterPlugin``.

    .. code-block:: json

        {
            "plugins": [
                {
                    "my_otter_plugin_package.MyOtterPlugin": {
                        "some_metadata_key": "some_value"
                    }
                },
                "my_otter_plugin_package.MyOtherOtterPlugin"
            ]
        }

    Args:
        submission_path (``str``): the absolute path to the submission being graded
        submission_metadata (``dict``): submission metadata; if on Gradescope, see
            https://gradescope-autograders.readthedocs.io/en/latest/submission_metadata/
        plugin_config (``dict``): configurations from the ``otter_config.json`` for this plugin, pulled
            from ``otter_config["plugins"][][PLUGIN_NAME]`` if ``otter_config["plugins"][]`` is a
            ``dict``

    Attributes:
        submission_path (``str``): the absolute path to the submission being graded
        submission_metadata (``dict``): submission metadata; if on Gradescope, see
            https://gradescope-autograders.readthedocs.io/en/latest/submission_metadata/
        plugin_config (``dict``): configurations from the ``otter_config.json`` for this plugin, pulled
            from ``otter_config["plugins"][][PLUGIN_NAME]`` if ``otter_config["plugins"][]`` is a
            ``dict``
    """

    def __init__(self, submission_path, submission_metadata, plugin_config):
        self.submission_path = submission_path
        self.submission_metadata = submission_metadata
        self.plugin_config = plugin_config

    def during_assign(self, assignment):
        """
        Plugin event run during the execution of Otter Assign after output directories are wrriten.
        Assignment configurations are passed in via the ``assignment`` argument.

        Args:
            assignment (``otter.assign.assignment.Assignment``): the ``Assignment`` instance with
                configurations for the assignment; used similar to an ``Attrdict`` where keys are
                accessed with the dot syntax (e.g. ``assignment.master`` is the path to the master
                notebook)

        Raises:
            ``PluginEventNotSupportedException``: if the event is not supported by this plugin
        """
        raise PluginEventNotSupportedException()

    def during_generate(self, otter_config, assignment):
        """
        Plugin event run during the execution of Otter Generate that can modify the configurations
        to be written to ``otter_config.json``.

        Args:
            otter_config (``dict``): the dictionary of Otter configurations to be written to
                ``otter_config.json`` in the zip file
            assignment (``otter.assign.assignment.Assignment``): the ``Assignment`` instance with
                configurations for the assignment if Otter Assign was used to generate this zip file;
                will be set to ``None`` if Otter Assign is not being used

        Raises:
            ``PluginEventNotSupportedException``: if the event is not supported by this plugin
        """
        raise PluginEventNotSupportedException()

    def from_notebook(self, *args, **kwargs):
        """
        Plugin event run by students as they work through a notebook via the ``Notebook`` API (see
        ``Notebook.run_plugin``). Accepts arbitrary arguments and has no return.

        Args:
            *args: arguments for the plugin event
            **kwargs: keyword arguments for the plugin event

        Raises:
            ``PluginEventNotSupportedException``: if the event is not supported by this plugin
        """
        raise PluginEventNotSupportedException()

    def notebook_export(self, *args, **kwargs):
        """
        Plugin event run when a student calls ``Notebook.export``. Accepts arbitrary arguments and
        should return a list of file paths to include in the exported zip file.

        Args:
            *args: arguments for the plugin event
            **kwargs: keyword arguments for the plugin event

        Returns:
            ``list[str]``: the list of file paths to include in the export

        Raises:
            ``PluginEventNotSupportedException``: if the event is not supported by this plugin
        """
        raise PluginEventNotSupportedException()

    def before_grading(self, config):
        """
        Plugin event run before the execution of the submission which can modify the dictionary of
        grading configurations.

        Args:
            config (``otter.run.run_autograder.autograder_config.AutograderConfig``): the
                autograder config

        Raises:
            ``PluginEventNotSupportedException``: if the event is not supported by this plugin
        """
        raise PluginEventNotSupportedException()

    def before_execution(self, submission):
        """
        Plugin event run before the execution of the submission which can modify the submission itself.
        This method should return a properly-formatted ``NotebookNode`` or string that will be executed in
        place of the student's original submission.

        Args:
            submission (``nbformat.NotebookNode`` or ``str``): the submission for grading; if it is
                a notebook, this will be the JSON-parsed ``dict`` of its contents; if it is a script,
                this will be a string containing the code

        Returns:
            ``nbformat.NotebookNode`` or ``str``: the altered submission to be executed

        Raises:
            ``PluginEventNotSupportedException``: if the event is not supported by this plugin
        """
        raise PluginEventNotSupportedException()

    def after_grading(self, results):
        """
        Plugin event run after all tests are run on the resulting environment that gets passed the
        ``otter.test_files.GradingResults`` object that stores the grading results for each test case.

        Args:
            results (``otter.test_files.GradingResults``): the results of all test cases

        Raises:
            ``PluginEventNotSupportedException``: if the event is not supported by this plugin
        """
        raise PluginEventNotSupportedException()

    def generate_report(self):
        """
        Plugin event run after grades are written to disk. This event should return a string that gets
        printed to stdout as a part of the plugin report.

        Returns:
            ``str``: the string to be included in the report

        Raises:
            ``PluginEventNotSupportedException``: if the event is not supported by this plugin
        """
        raise PluginEventNotSupportedException()



"""Utilities for Otter Assign"""

import copy
import json
import nbformat as nbf
import os
import pathlib
import re
import shutil
import warnings

from textwrap import indent
from typing import Any, Optional, TYPE_CHECKING

from .. import logging
from ..api import grade_submission
from ..generate import main as generate_autograder
from ..plugins import PluginCollection
from ..run import capture_run_output
from ..utils import chdir, get_source, OTTER_CONFIG_FILENAME


LOGGER = logging.get_logger(__name__)


class EmptyCellException(Exception):
    """
    Exception for empty cells to indicate deletion
    """


class AssignNotebookFormatException(Exception):
    """
    An exception representing an error in the formatting of an Otter Assign master notebook.

    Wraps an error message with information about the question and cell to make the location of the
    error easy to find for users.

    Args:
        message (``str``): the error message
        question (``otter.assign.question_config.QuestionConfig | None``): the question config if
            the error occurred inside of a question block
        cell_index (``int``): the index of the cell
        *args: additional args passed to ``Exception`` constructor
        **kwargs: additional keyword args passed to ``Exception`` constructor
    """

    def __init__(
        self,
        message: str,
        question: Optional["QuestionConfig"],
        cell_index: int,
        *args: Any,
        **kwargs: Any,
    ):
        message += " ("
        if question is not None:
            message += f"question { question.name }, "
        message += f"cell number { cell_index + 1 })"
        super().__init__(message, *args, **kwargs)


def get_notebook_language(nb: nbf.NotebookNode) -> str:
    """
    Parse the notebook kernel language and return it as a string.

    Args:
        nb (``nbformat.NotebookNode``): the notebook

    Returns:
        ``str``: the name of the language as a lowercased string

    Warns:
        if there is no ``language`` key in the notebook's kernelspec
    """
    kernelspec = nb["metadata"]["kernelspec"]
    if "language" in kernelspec:
        return kernelspec["language"].lower()
    warnings.warn("Could not auto-parse kernelspec from notebook; assuming Python")
    return "python"


def is_ignore_cell(cell: nbf.NotebookNode) -> bool:
    """
    Returns whether the current cell should be ignored

    Args:
        cell (``nbformat.NotebookNode``): a notebook cell

    Returns:
        ``bool``: whether the cell is a ignored
    """
    source = get_source(cell)
    return bool(
        source and re.match(r"(##\s*ignore\s*##\s*|#\s*ignore\s*)", source[0], flags=re.IGNORECASE)
    )


def is_cell_type(cell: nbf.NotebookNode, cell_type: str) -> bool:
    """
    Determine whether a cell is of a specific type.

    Args:
        cell (``nbformat.NotebookNode``): the cell to check
        cell_type (``str``): the cell type to check for

    Returns:
        ``bool``: whether ``cell`` is of type ``cell_type``
    """
    return cell["cell_type"] == cell_type


def remove_output(nb: nbf.NotebookNode) -> None:
    """
    Remove all outputs from a notebook in-place.

    Args:
        nb (``nbformat.NotebookNode``): a notebook
    """
    for cell in nb["cells"]:
        if "outputs" in cell:
            cell["outputs"] = []
        if "execution_count" in cell:
            cell["execution_count"] = None


def lock(cell: nbf.NotebookNode) -> None:
    """
    Make a cell non-editable and non-deletable in-place.

    Args:
        cell (``nbformat.NotebookNode``): cell to be locked
    """
    m = cell["metadata"]
    m["editable"] = False
    m["deletable"] = False


def add_tag(cell: nbf.NotebookNode, tag: str) -> nbf.NotebookNode:
    """
    Add a tag to a cell, returning a copy.

    Args:
        cell (``nbformat.NotebookNode``): the cell to add a tag to
        tag (``str``): the tag to add

    Returns:
        ``nbformat.NotebookNode``: a copy of ``cell`` with the tag added
    """
    cell = copy.deepcopy(cell)
    if "tags" not in cell["metadata"]:
        cell["metadata"]["tags"] = []
    cell["metadata"]["tags"].append(tag)
    return cell


def has_tag(cell: nbf.NotebookNode, tag: str) -> bool:
    """
    Determine whether a cell has a tag.

    Args:
        cell (``nbformat.NotebookNode``): the cell to check
        tag (``str``): the tag to check for

    Returns:
        ``nbformat.NotebookNode``: whether ``cell`` is tagged with ``tag``
    """
    return tag in cell["metadata"].get("tags", [])


def remove_tag(cell: nbf.NotebookNode, tag: str) -> nbf.NotebookNode:
    """
    Remove a tag from a cell, returning a copy.

    Args:
        cell (``nbformat.NotebookNode``): the cell to remove a tag from
        tag (``str``): the tag to remove

    Returns:
        ``nbformat.NotebookNode``: a copy of ``cell`` with the tag removed
    """
    cell = copy.deepcopy(cell)
    if "tags" not in cell["metadata"] or tag not in cell["metadata"]["tags"]:
        return cell
    cell["metadata"]["tags"].remove(tag)
    return cell


def str_to_doctest(code_lines: list[str], lines: list[str]) -> list[str]:
    """
    Convert a list of lines of Python code ``code_lines`` to the doctest format and appending the
    results to ``lines``.

    Args:
        code_lines (``list[str]``): the code to convert
        lines (``list[str]``): the list to append the converted lines to

    Returns:
        ``list[str]``: a pointer to ``lines``
    """
    if len(code_lines) == 0:
        return lines
    line = code_lines.pop(0)
    # skip empty lines
    while not line:
        line = code_lines.pop(0)
    if line.startswith(" ") or line.startswith("\t"):
        return str_to_doctest(code_lines, lines + ["... " + line])
    elif (
        bool(re.match(r"^except[\s\w]*:", line))
        or line.startswith("elif ")
        or line.startswith("else:")
        or line.startswith("finally:")
    ):
        return str_to_doctest(code_lines, lines + ["... " + line])
    elif len(lines) > 0 and lines[-1].strip().endswith("\\"):
        return str_to_doctest(code_lines, lines + ["... " + line])
    else:
        return str_to_doctest(code_lines, lines + [">>> " + line])


def run_tests(assignment: "Assignment", debug: bool = False) -> None:
    """
    Grade a notebook and throw an error if it does not receive a perfect score.

    Args:
        assignment (``otter.assign.assignment.Assignment``): the assignment config
        debug (``bool``): whether to throw errors instead of swallowing them during grading

    Raises:
        ``RuntimeError``: if the grade received by the notebook is not 100%
    """
    with capture_run_output() as run_output:
        results = grade_submission(
            str(assignment.ag_notebook_path),
            str(assignment.ag_zip_path),
            debug=debug,
            extra_submission_files=assignment.student_files,
        )

    LOGGER.debug(f"Otter Run output:\n{run_output.getvalue()}")

    if results.total != results.possible:
        raise RuntimeError(
            f"Some autograder tests failed in the autograder notebook:\n"
            + indent(results.summary(), "    ")
        )


def write_otter_config_file(assignment: "Assignment") -> None:
    """
    Create an Otter configuration file (a ``.otter`` file) for students to use Otter tools,
    including saving environments and submitting to an Otter Service deployment, using assignment
    configurations. Writes the resulting file to the ``autograder`` and ``student`` subdirectories
    of ``assignment.result``.

    Args:
        assignment (``otter.assign.assignment.Assignment``): the assignment config
    """
    config = {}

    config["notebook"] = assignment.master.name
    config["save_environment"] = assignment.save_environment
    config["ignore_modules"] = assignment.ignore_modules

    if assignment.generate and assignment.generate.serialized_variables:
        config["variables"] = assignment.generate.serialized_variables

    config_name = assignment.master.stem + ".otter"
    with open(assignment.get_ag_path(config_name), "w+") as f:
        json.dump(config, f, indent=4)
    with open(assignment.get_stu_path(config_name), "w+") as f:
        json.dump(config, f, indent=4)


def run_generate_autograder(
    assignment: "Assignment",
    gs_username: Optional[str],
    gs_password: Optional[str],
    plugin_collection: Optional[PluginCollection] = None,
) -> None:
    """
    Run Otter Generate on the autograder directory to generate a Gradescope zip file. Relies on
    configurations in ``assignment.generate``.

    Args:
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
        gs_username (``str | None``): Gradescope username for token generation
        gs_password (``str | None``): Gradescope password for token generation
        plugin_collection (``otter.plugins.PluginCollection``): a plugin collection to pass
            to Otter Generate
    """
    curr_dir = os.getcwd()
    with chdir(str(assignment.get_ag_path())):

        # use temp tests dir
        test_dir = "tests"
        if (
            assignment.is_python
            and not assignment.tests.files
            and assignment.generate_tests_dir is None
        ):
            raise RuntimeError("Failed to create temp tests directory for Otter Generate")

        elif assignment.is_python and not assignment.tests.files:
            test_dir = assignment.generate_tests_dir

        requirements = None
        if assignment.requirements:
            if assignment.is_r:
                requirements = "requirements.R"
            else:
                requirements = "requirements.txt"

        files = []
        if assignment.files:
            files += assignment.files

        if assignment.autograder_files:
            with chdir(curr_dir):
                output_dir = assignment.get_ag_path()

                # copy files
                for file in assignment.autograder_files:

                    # if a directory, copy the entire dir
                    if os.path.isdir(file):
                        shutil.copytree(file, str(output_dir / os.path.basename(file)))

                    else:
                        # check that file is in subdir
                        assert os.getcwd() in os.path.abspath(
                            file
                        ), f"{file} is not in a subdirectory of the master notebook directory"
                        file_path = pathlib.Path(file).resolve()
                        rel_path = file_path.parent.relative_to(pathlib.Path(os.getcwd()))
                        os.makedirs(output_dir / rel_path, exist_ok=True)
                        shutil.copy(file, str(output_dir / rel_path))

            files += assignment.autograder_files

        otter_config = assignment.get_otter_config()
        if otter_config:
            with open(OTTER_CONFIG_FILENAME, "w+") as f:
                json.dump(otter_config, f, indent=2)

        generate_autograder(
            tests_dir=test_dir,
            output_path=assignment.ag_zip_name,
            config=OTTER_CONFIG_FILENAME if otter_config else None,
            lang="python" if assignment.is_python else "r",
            requirements=requirements,
            overwrite_requirements=assignment.overwrite_requirements,
            environment="environment.yml" if assignment.environment else None,
            no_environment=False,
            username=gs_username,
            password=gs_password,
            files=files,
            plugin_collection=plugin_collection,
            assignment=assignment,
            python_version=assignment.get_python_version(),
            channel_priority_strict=assignment.channel_priority_strict,
            exclude_conda_defaults=assignment.exclude_conda_defaults,
        )

        # clean up temp tests dir
        if assignment.generate_tests_dir is not None:
            shutil.rmtree(assignment.generate_tests_dir)


def remove_cell_ids_if_applicable(nb: nbf.NotebookNode) -> None:
    """
    Remove all cell IDs from a notebook in-place iff the nbformat of the notebook is < 4.5.

    Args:
        nb (``nbformat.NotebookNode``): a notebook
    """
    if nb["nbformat"] < 4 or (nb["nbformat"] == 4 and nb["nbformat_minor"] < 5):
        for cell in nb.cells:
            if "id" in cell:
                cell.pop("id")


if TYPE_CHECKING:
    from .assignment import Assignment
    from .question_config import QuestionConfig



"""Argument parser for Otter command-line tools"""

import click
import functools

from typing import Any, Callable

from . import logging
from .assign import main as assign
from .check import main as check
from .export import main as export
from .generate import main as generate
from .grade import ALLOWED_EXTENSIONS, main as grade
from .run import main as run
from .version import print_version_info


_VERBOSITY_LEVELS = {
    0: logging.WARNING,
    1: logging.INFO,
    2: logging.DEBUG,
}


def _verbosity(f: Callable[..., Any]):
    @click.option("-v", "--verbose", "verbosity", count=True, help="Verbosity of the logged output")
    @functools.wraps(f)
    def wrapper(*args: Any, **kwargs: Any):
        # set the log level
        verbosity = kwargs.pop("verbosity")
        logging.set_level(_VERBOSITY_LEVELS[min(verbosity, max(_VERBOSITY_LEVELS.keys()))])
        return f(*args, **kwargs)

    return wrapper


@click.group(invoke_without_command=True)
@click.option("--version", is_flag=True, help="Show the version and exit")
def cli(version: bool):
    """
    Command-line utility for Otter-Grader, a Python-based autograder for Jupyter Notebooks,
    RMarkdown files, and Python and R scripts. For more information, see
    https://otter-grader.readthedocs.io/.
    """
    if version:
        print_version_info(logo=True)
        return


defaults = assign.__kwdefaults__


@cli.command("assign")
@_verbosity
@click.argument("master", type=click.Path(exists=True, dir_okay=False))
@click.argument("result", type=click.Path())
@click.option(
    "--no-run-tests", is_flag=True, help="Do not run the tests against the autograder notebook"
)
@click.option("--no-pdfs", is_flag=True, help="Do not generate PDFs; overrides assignment config")
@click.option("--username", help="Gradescope username for generating a token")
@click.option("--password", help="Gradescope password for generating a token")
@click.option("--debug", is_flag=True, help="Do not ignore errors in running tests for debugging")
def assign_cli(*args: Any, **kwargs: Any):
    """
    Create distribution versions of the Otter Assign formatted notebook MASTER and write the
    results to the directory RESULT, which will be created if it does not already exist.
    """
    return assign(*args, **kwargs)


defaults = check.__kwdefaults__


@cli.command("check")
@_verbosity
@click.argument("file", type=click.Path(exists=True, dir_okay=False))
@click.option("-q", "--question", help="A specific question to grade")
@click.option(
    "-t",
    "--tests-path",
    default=defaults["tests_path"],
    type=click.Path(exists=True, file_okay=False),
    help="Path to the directory of test files",
)
@click.option("--seed", type=click.INT, help="A random seed to be executed before each cell")
def check_cli(*args: Any, **kwargs: Any):
    """
    Check the Python script or Jupyter Notebook FILE against tests.
    """
    return check(*args, **kwargs)


defaults = export.__kwdefaults__


@cli.command("export")
@_verbosity
@click.argument("src", type=click.Path(exists=True, dir_okay=False))
@click.argument("dest", required=False, type=click.Path())
@click.option("--filtering", is_flag=True, help="Whether the PDF should be filtered")
@click.option(
    "--pagebreaks", is_flag=True, help="Whether the PDF should have pagebreaks between questions"
)
@click.option("-s", "--save", is_flag=True, help="Save intermediate file(s) as well")
@click.option(
    "-e",
    "--exporter",
    default=defaults["exporter"],
    type=click.Choice(["latex", "html"]),
    help="Type of PDF exporter to use",
)
@click.option("--xecjk", is_flag=True, help="Enable xeCJK in Otter's LaTeX template")
def export_cli(*args: Any, **kwargs: Any):
    """
    Export a Jupyter Notebook SRC as a PDF at DEST with optional filtering.

    If unspecified, DEST is assumed to be the basename of SRC with a .pdf extension.
    """
    return export(*args, **kwargs)


defaults = generate.__kwdefaults__


@cli.command("generate")
@_verbosity
@click.option(
    "-t",
    "--tests-dir",
    default=defaults["tests_dir"],
    type=click.Path(exists=True, file_okay=False),
    help="Path to test files",
)
@click.option(
    "-o",
    "--output-path",
    default=defaults["output_path"],
    type=click.Path(),
    help="Path at which to write autograder zip file",
)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to otter configuration file; ./otter_config.json automatically checked",
)
@click.option(
    "--no-config",
    is_flag=True,
    help="Disable auto-inclusion of unspecified Otter config file at ./otter_config.json",
)
@click.option(
    "-r",
    "--requirements",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to requirements.txt file; ./requirements.txt automatically checked",
)
@click.option(
    "--no-requirements",
    is_flag=True,
    help="Disable auto-inclusion of unspecified requirements file at ./requirements.txt",
)
@click.option(
    "--overwrite-requirements",
    is_flag=True,
    help="Overwrite (rather than append to) default requirements for Gradescope; ignored if no REQUIREMENTS argument",
)
@click.option(
    "-e",
    "--environment",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to environment.yml file; ./environment.yml automatically checked (overwrite)",
)
@click.option(
    "--no-environment",
    is_flag=True,
    help="Disable auto-inclusion of unspecified environment file at ./environment.yml",
)
@click.option("-l", "--lang", help="Assignment programming language; defaults to Python")
@click.option("--username", help="Gradescope username for generating a token")
@click.option("--password", help="Gradescope password for generating a token")
@click.option("--token", help="Gradescope token for uploading PDFs")
@click.option("--python-version", help="Python version to use in the grading image")
@click.option(
    "--channel-priority-strict",
    is_flag=True,
    help="Whether to set conda's channel_priority to strict in the setup.sh file",
)
@click.option(
    "--exclude-conda-defaults",
    is_flag=True,
    help="Whether to exclude conda's defaults channel from the environment.yml file",
)
@click.argument("files", nargs=-1)
def generate_cli(*args: Any, **kwargs: Any):
    """
    Generate a zip file to configure an Otter autograder, including FILES as support files.
    """
    return generate(*args, **kwargs)


defaults = grade.__kwdefaults__


@cli.command("grade")
@_verbosity
@click.argument("paths", nargs=-1, type=click.Path(exists=True))
@click.option("-n", "--name", help="An assignment name to use in the Docker image tag")
@click.option(
    "-a", "--autograder", default=defaults["autograder"], help="Path to autograder zip file"
)
@click.option(
    "-o", "--output-dir", default=defaults["output_dir"], help="Directory to which to write output"
)
@click.option(
    "--ext",
    default=defaults["ext"],
    type=click.Choice(ALLOWED_EXTENSIONS),
    help="The extension to glob for submissions",
)
@click.option(
    "--summaries",
    is_flag=True,
    help="Whether to write the otter run results for each graded notebook",
)
@click.option("--pdfs", is_flag=True, help="Whether to copy notebook PDFs out of containers")
@click.option(
    "--containers",
    default=defaults["containers"],
    type=click.INT,
    help="Specify number of containers to run in parallel",
)
@click.option(
    "--image", default=defaults["image"], help="A Docker image tag to use as the base image"
)
@click.option("--timeout", type=click.INT, help="Submission execution timeout in seconds")
@click.option("--no-network", is_flag=True, help="Disable networking in the containers")
@click.option("--no-kill", is_flag=True, help="Do not kill containers after grading")
@click.option(
    "--debug",
    is_flag=True,
    help="Run in debug mode (without ignoring errors thrown during execution)",
)
@click.option("--prune", is_flag=True, help="Prune all of Otter's grading images")
@click.option("-f", "--force", is_flag=True, help="Force action (don't ask for confirmation)")
def grade_cli(*args: Any, **kwargs: Any):
    """
    Grade submissions in PATHS locally using Docker containers. PATHS can be individual file paths
    or directories containing submissions ending with extension EXT.
    """
    kwargs["result_queue"] = None
    g = grade(*args, **kwargs)
    if g is not None:
        click.echo(g)
    return g


defaults = run.__kwdefaults__


@cli.command("run")
@_verbosity
@click.argument("submission")
@click.option(
    "-a",
    "--autograder",
    default=defaults["autograder"],
    type=click.Path(exists=True, dir_okay=False),
    help="Path to autograder zip file",
)
@click.option(
    "-o",
    "--output-dir",
    default=defaults["output_dir"],
    type=click.Path(exists=True, file_okay=False),
    help="Directory to which to write output",
)
@click.option("--no-logo", is_flag=True, help="Suppress Otter logo in stdout")
@click.option("--debug", is_flag=True, help="Do not ignore errors when running submission")
@click.option("-p", "--pickle-results", is_flag=True, help="Output GradingResults pickle file")
def run_cli(*args: Any, **kwargs: Any):
    """
    Run non-containerized Otter on a single submission, writing results to a JSON file.
    """
    write_pkl = kwargs.pop("pickle_results", False)
    results = run(*args, **kwargs)
    if write_pkl:
        import dill

        with open("results.pkl", "wb+") as f:
            dill.dump(results, f)
    return results



"""Class for running tests from test files"""

import inspect

from typing import Any, ClassVar, Optional

from ..nbmeta_config import NBMetadataConfig
from ..test_files import create_test_file, TestFile


class Checker:
    """
    A class for running and optionally tracking checks against test files.

    This class is not meant to be instantiated and is composed solely of class methods.
    """

    _track_results: ClassVar[bool] = False
    """whether to store check results"""

    _test_files: ClassVar[list[TestFile]] = []
    """stored check results"""

    def __new__(cls, *args: Any, **kwargs: Any):
        raise NotImplementedError("The Checker class cannot be instantiated")

    @classmethod
    def enable_tracking(cls):
        """
        Enable the tracking of test results from calls to ``check``.
        """
        cls._track_results = True

    @classmethod
    def disable_tracking(cls):
        """
        Disable the tracking of test results from calls to ``check``.
        """
        cls._track_results = False

    @classmethod
    def get_results(cls) -> list[TestFile]:
        """
        Get a pointer to the list into which check results are being collected.
        """
        return cls._test_files

    @classmethod
    def clear_results(cls):
        """
        Overwrite the pointer to the check result collection list.

        Does not affect the original list, only overwrites the field in the ``Checker`` class that
        points to it.
        """
        cls._test_files = []

    @classmethod
    def check(
        cls,
        nb_or_test_path: str,
        nbmeta_config: NBMetadataConfig,
        test_name: Optional[str] = None,
        global_env: Optional[dict[str, Any]] = None,
    ) -> TestFile:
        """
        Checks a global environment against a test, which may be stored in a file or in a notebook's
        metadata.

        Args:
            nb_or_test_path (``str``): path to test file or notebook
            test_name (``str``): the name of the test if a notebook metadata test
            global_env (``dict``): the global environment in which to run the test; if
                unspecified, the calling frame's global environment is used

        Returns:
            ``otter.test_files.abstract_test.TestFile``: result of running the tests in the
            given global environment
        """
        test = create_test_file(nb_or_test_path, nbmeta_config, test_name=test_name)

        if global_env is None:
            global_env = inspect.currentframe().f_back.f_globals

        test.run(global_env)

        if cls._track_results:
            cls._test_files.append(test)

        return test

    @classmethod
    def check_if_not_already_checked(
        cls, test_path: str, global_env: Optional[dict[str, Any]] = None
    ) -> Optional[TestFile]:
        """
        Run the specified test if it has not already been run (that is, if its result is not cached
        in this ``Checker`` instance).

        This method only works with test files.

        Args:
            test_path (``str``): path to test file
            global_env (``dict[str, Any] | None``): the global environment in which to run the test; if
                unspecified, the calling frame's global environment is used

        Returns:
            ``otter.test_files.abstract_test.TestFile | None``: result of running the tests in the
            given global environment
        """
        if any(test_path in tf.path or tf.path in test_path for tf in cls._test_files):
            return

        if global_env is None:
            global_env = inspect.currentframe().f_back.f_globals

        return cls.check(test_path, NBMetadataConfig(), global_env=global_env)



"""Abstract runner class for the autograder"""

import gc
import json
import nbformat
import os
import shutil
import tempfile

from abc import ABC, abstractmethod
from glob import glob
from typing import Optional

from ..autograder_config import AutograderConfig
from ..utils import OtterRuntimeError, print_output, write_blank_page_to_stare_at_before_you
from ....generate.token import APIClient
from ....nbmeta_config import NBMetadataConfig
from ....test_files import GradingResults


class AbstractLanguageRunner(ABC):
    """
    A class defining the logic of running the autograder and generating grading results.

    Args:
        otter_config (``dict[str, object]``): user-specified configurations to override the defaults
        **kwargs: other user-specified configurations to override the defaults and values specified
            in ``otter_config``
    """

    ag_config: AutograderConfig
    """the autograder config"""

    def __init__(self, ag_config: AutograderConfig):
        self.ag_config = ag_config

    def prepare_files(self):
        """
        Copies tests and support files needed for running the autograder.

        When this method is invoked, the working directory is assumed to already be
        ``self.ag_config.autograder_dir``.
        """
        # put files into submission directory
        if os.path.exists("./source/files"):
            for file in os.listdir("./source/files"):
                fp = os.path.join("./source/files", file)
                if os.path.isdir(fp):
                    if not os.path.exists(os.path.join("./submission", os.path.basename(fp))):
                        shutil.copytree(fp, os.path.join("./submission", os.path.basename(fp)))
                else:
                    shutil.copy(fp, "./submission")

        # copy the tests directory
        if os.path.exists("./submission/tests"):
            shutil.rmtree("./submission/tests")
        shutil.copytree("./source/tests", "./submission/tests")

    def validate_assignment_name(self, got: Optional[str]):
        """
        Raise an ``OtterRuntimeError`` if the assignment name of the notebook is invalid.

        If no assignment name was provided in the configurations, no action is taken.

        Args:
            got (``str | None``): the assignment name of the submission or ``None`` if it didn't
                have one

        Raises:
            ``otter.run.run_autograder.utils.OtterRuntimeError``: if the name is invalid and
                grading should be aborted
        """
        if self.ag_config.assignment_name and got != self.ag_config.assignment_name:
            message = (
                f"Received submission for assignment '{got}' (this is assignment "
                f"'{self.ag_config.assignment_name}')"
            )
            raise OtterRuntimeError(message)

    def get_notebook_assignment_name(self, nb: nbformat.NotebookNode) -> Optional[str]:
        """
        Get the assignment name in the metadata of the provided notebook, if any.

        Args:
            nb (``nbformat.NotebookNode``): the notebook

        Returns:
            ``str | None``: the assignment name of the notebook, if any
        """
        nbmc = NBMetadataConfig.from_notebook(nb)
        return nbmc.assignment_name

    def write_and_maybe_submit_pdf(self, submission_path: str) -> Optional[Exception]:
        """
        Upload a PDF to a Gradescope assignment for manual grading.

        This method calls ``self.write_pdf`` to generate the PDF before submitting it, and catches
        any errors thrown in that function.

        This method should be called, if appropriate, by ``self.run``.

        Args:
            submission_path (``str``): path to the submission

        Returns:
            ``Exception | None``: an error thrown while trying to write or submit the PDF, if any;
                a return value of ``None`` indicates a successful export and upload
        """
        # Don't export or submit a PDF in debug mode.
        if self.ag_config.debug:
            return None

        try:
            subm_pdfs = glob("*.pdf")
            if self.ag_config.use_submission_pdf and subm_pdfs:
                pdf_path = subm_pdfs[0]
            else:
                pdf_path = self.write_pdf(submission_path)

            if self.ag_config.token:
                client = APIClient(token=self.ag_config.token)
                self.submit_pdf(client, pdf_path)

                # ensure the client gets garbage collected so that the token can't be accessed
                # by student code
                del client
                gc.collect()

        except Exception as e:
            print_output(f"\n\nError encountered while generating and submitting PDF:\n{e}")

            if self.ag_config.submit_blank_pdf_on_export_failure and self.ag_config.token:
                print_output("\nUploading a blank PDF due to export failure")
                with tempfile.NamedTemporaryFile(suffix=".pdf") as ntf:
                    write_blank_page_to_stare_at_before_you(ntf.name)
                    self.submit_pdf(client, ntf.name)

            return e

    def submit_pdf(self, client: APIClient, pdf_path: str):
        """
        Upload the PDF at ``pdf_path`` to the Gradescope assignment specified in the config. Does
        not check whether the assignment configuration is valid.

        Args:
            client (``otter.generate.token.APIClient``): the Gradescope client
            pdf_path (``str``): path to the PDF file to upload
        """
        # get student email
        with open("../submission_metadata.json", encoding="utf-8") as f:
            metadata = json.load(f)

        student_emails = []
        for user in metadata.get("users", []):
            student_emails.append(user["email"])

        # validate that a course and assignment ID were specified
        if self.ag_config.course_id is None:
            raise OtterRuntimeError("PDF upload course ID not specified")
        if self.ag_config.assignment_id is None:
            raise OtterRuntimeError("PDF upload assignment ID not specified")

        for student_email in student_emails:
            res = client.upload_pdf_submission(
                self.ag_config.course_id,
                self.ag_config.assignment_id,
                student_email,
                pdf_path,
            )
            if res.status_code != 200:
                raise OtterRuntimeError(
                    f"Failed to upload submission for {student_email}: [status {res.status_code}] {res.text}"
                )

        print_output(
            "\n\nSuccessfully uploaded submissions for: {}".format(", ".join(student_emails))
        )

    def sanitize_tokens(self):
        """
        Sanitize any references to the PDF submission upload token to prevent unauthorized access
        when executing student code.

        This method should be invoked from inside ``{self.ag_config.autograder_dir}/submission``
        as part of ``run``.
        """
        self.ag_config.token = None
        if not os.path.exists("../source/otter_config.json"):
            return
        with open("../source/otter_config.json") as f:
            c = json.load(f)
        if "token" in c:
            del c["token"]
        with open("../source/otter_config.json", "w") as f:
            json.dump(c, f, indent=2)

    @abstractmethod
    def validate_submission(self, submission_path: str):
        """
        Validate the submission, raising an error/warning if appropriate.

        This method should be invoked as part of the implementation of another abstract method
        (either ``resolve_submission_path`` or ``run``).

        Args:
            submission_path (``str``): the path to the submission file
        """
        ...

    @abstractmethod
    def resolve_submission_path(self) -> str:
        """
        Determine the path to the submission file, performing any necessary transformations on the
        file.

        When this method is invoked, the working directory is assumed to already be
        ``{self.ag_config.autograder_dir}/submission``.
        """
        ...

    @abstractmethod
    def write_pdf(self, submission_path: str) -> str:
        """
        Convert the submission to a PDF, returning the path to the PDF file.

        Args:
            submission_path (``str``): the path to the submission

        Returns:
            ``str``: the path to the generated PDF
        """
        ...

    @abstractmethod
    def run(self) -> GradingResults:
        """
        Run the autograder according to the configurations in ``self.ag_config``.

        When this method is invoked, the working directory is assumed to already be
        ``self.ag_config.autograder_dir``.

        Returns:
            ``otter.test_files.GradingResults``: the results from grading the submission
        """
        ...
